"""Language model baselines in TensorFlow
"""
from itertools import chain
from baseline.tf.tfy import *
from baseline.version import __version__
from baseline.model import LanguageModel, register_model
from baseline.tf.embeddings import *
from baseline.tf.tfy import new_placeholder_dict, TRAIN_FLAG, lstm_cell_w_dropout
from baseline.utils import read_json, write_json, MAGIC_VARS


class LanguageModelBase(tf.keras.Model, LanguageModel):
    """Base for all baseline implementations of LMs

    This class provides a loose skeleton around which the baseline models
    are built.  This essentially consists of dividing up the network into a logical separation between "embedding",
    or composition of lookup tables to build a vector representation of a temporal input, "decoding",
    or the conversion of temporal data to a decoded representation, and "output" --
    a projection to output space and a softmax
    """
    def __init__(self):
        """Construct a base LM
        """
        super().__init__()
        self.saver = None
        self.hsz = None
        self.probs = None
        self._unserializable = []

    def save_values(self, basename):
        """Save tensor files out

        :param basename: Base name of model
        :return:
        """
        if not tf.executing_eagerly():
            self.saver.save(self.sess, basename, write_meta_graph=False)
        else:
            self.save_weights(f"{basename}.wgt")

    def save_md(self, basename):
        """This method saves out a `.state` file containing meta-data from these classes and any info
        registered by a user-defined derived class as a `property`. Also write the `graph` and `saver` and `labels`

        :param basename:
        :return:
        """

        write_json(self._state, basename + '.state')
        for key, embedding in self.embeddings.items():
            embedding.save_md(basename + '-{}-md.json'.format(key))

    def _record_state(self, embeddings, **kwargs):
        """
        First, write out the embedding names, so we can recover those.  Then do a deepcopy on the model init params
        so that it can be recreated later.  Anything that is a placeholder directly on this model needs to be removed

        :param kwargs:
        :return:
        """
        embeddings_info = {}
        for k, v in embeddings.items():
            embeddings_info[k] = v.__class__.__name__

        blacklist = set(chain(self._unserializable, MAGIC_VARS, embeddings.keys()))
        self._state = {k: v for k, v in kwargs.items() if k not in blacklist}
        self._state.update({
            'version': __version__,
            'module': self.__class__.__module__,
            'class': self.__class__.__name__,
            'embeddings': embeddings_info,
        })

    def set_saver(self, saver):
        """Connect a `tf.Saver` to the model

        :param saver: A saver
        :return: None
        """
        self.saver = saver

    def save(self, basename):
        """Save the model

        :param basename: The model prefix
        :return:
        """
        self.save_md(basename)
        self.save_values(basename)

    def _create_loss(self, scope):
        with tf.compat.v1.variable_scope(scope):
            targets = tf.reshape(self.y, [-1])
            bt_x_v = tf.nn.log_softmax(tf.reshape(self.logits, [-1, self.vsz]), axis=-1)
            one_hots = tf.one_hot(targets, self.vsz)
            example_loss = -tf.reduce_sum(one_hots * bt_x_v, axis=-1)
            loss = tf.reduce_mean(example_loss)
            return loss

    def create_loss(self):
        """Create training loss operator

        :return: loss
        """
        return self._create_loss(scope='loss')

    def create_test_loss(self):
        """Create test loss operator

        :return: loss
        """
        return self._create_loss(scope='test_loss')

    def make_input(self, batch_dict, train=False):
        """When we are running with `DataFeed`s, need to transform to `feed_dict`s

        :param batch_dict: The batch for a step
        :param train: (`bool`) Are we training (or evaluating)?
        :return: A `feed_dict`
        """
        if not tf.executing_eagerly():
            batch_dict_for_model = new_placeholder_dict(train)

            for key in self.src_keys:
                batch_dict_for_model["{}:0".format(key)] = batch_dict[key]

            y = batch_dict.get('y')
            if y is not None:
                batch_dict_for_model[self.y] = batch_dict['y']

        else:
            SET_TRAIN_FLAG(train)

            batch_dict_for_model = {}
            for key in self.src_keys:
                batch_dict_for_model[key] = batch_dict[key]

        return batch_dict_for_model

    def predict(self, batch_dict):
        """Do prediction from a `batch_dict`

        :param batch_dict: A step of data
        :return: The softmax output for this step
        """
        batch_dict = self.make_input(batch_dict)
        if not tf.executing_eagerly():
            step_softmax = self.sess.run(self.probs, batch_dict)
        else:
            # FIXME: This is not really the proper handling for eager mode
            # We want to be able to pass in the last hidden state and emit the current one right?
            step_softmax = tf.nn.softmax(self(batch_dict, None)[0])

        return step_softmax

    @classmethod
    def create(cls, embeddings, **kwargs):
        """Create the language model

        :param embeddings: A set of embeddings used
        :param kwargs: see below

        :Keyword Arguments:

        * *tgt_key* (`str`) -- Which vocabulary is the destination vocabulary
          (for example, you might have character inputs, or character + word inputs.  The outputs need to be specified)
        * *sess* (`tf.compat.v1.Session`) -- Optionally, pass in a session (or one will be created)
        * *pdrop* (`float`) -- The dropout probability
        * *y* -- Optional target.  If this is not passed in, a placeholder gets created
        * *hsz* (`int`) -- Number of hidden units per layers
        * *unif* (`float`) -- set the weights initializer to small random uniform values

        :return: The created model
        """
        lm = cls()
        lm.src_keys = kwargs.get('src_keys', embeddings.keys())
        lm.tgt_key = kwargs.get('tgt_key')
        if lm.tgt_key is None:
            raise Exception('Need a `tgt_key` to know which source vocabulary should be used for destination')

        lm._unserializable.append(lm.tgt_key)
        lm._record_state(embeddings, **kwargs)
        inputs = {}

        if not tf.executing_eagerly():

            for k, embedding in embeddings.items():
                x = kwargs.get(k, embedding.create_placeholder(name=k))
                inputs[k] = x

            lm.y = kwargs.get('y', tf.compat.v1.placeholder(tf.int32, [None, None], name="y"))
            lm.sess = kwargs.get('sess', tf.compat.v1.Session())
        lm.create_layers(embeddings, **kwargs)

        if not tf.executing_eagerly():
            if lm.requires_state:
                lm.zero_state(inputs)
                lm.logits, lm.final_state = lm(inputs, lm.initial_state)
            else:
                lm.logits, _ = lm(inputs, None)
            lm.probs = tf.nn.softmax(lm.logits, name="softmax")

        return lm

    def call(self, inputs: Dict[str, TensorDef], hidden: TensorDef) -> Tuple[TensorDef, TensorDef]:
        """Take the input and produce the best path of labels out

        :param inputs: The feature indices for the input
        :return: The output and hidden units
        """

    def create_layers(self, embeddings, **kwargs):
        """This method defines the model itself, and must be overloaded by derived classes

        This function will update `self` with the layers required to execute the `call()` method

        :param embeddings: The input feature indices
        :param kwargs:
        :return:
        """

    @classmethod
    def load(cls, basename, **kwargs):
        """Reload the model from a graph file and a checkpoint

        The model that is loaded is independent of the pooling and stacking layers, making this class reusable
        by sub-classes.

        :param basename: The base directory to load from
        :param kwargs: See below

        :Keyword Arguments:
        * *sess* -- An optional tensorflow session.  If not passed, a new session is
            created

        :return: A restored model
        """
        _state = read_json(basename + '.state')

        if not tf.executing_eagerly():
            _state['sess'] = kwargs.pop('sess', create_session())
            embeddings_info = _state.pop("embeddings")

            with _state['sess'].graph.as_default():
                embeddings = reload_embeddings(embeddings_info, basename)
                for k in embeddings_info:
                    if k in kwargs:
                        _state[k] = kwargs[k]

                _state['model_type'] = kwargs.get('model_type', 'default')
                model = cls.create(embeddings, **_state)
                model._state = _state

                do_init = kwargs.get('init', True)
                if do_init:
                    init = tf.compat.v1.global_variables_initializer()
                    model.sess.run(init)

                model.saver = tf.compat.v1.train.Saver()
                model.saver.restore(model.sess, basename)
        else:
            _state = read_json(basename + '.state')
            _state['model_type'] = kwargs.get('model_type', 'default')
            embeddings = {}
            embeddings_dict = _state.pop("embeddings")

            for key, class_name in embeddings_dict.items():
                md = read_json('{}-{}-md.json'.format(basename, key))
                embed_args = dict({'vsz': md['vsz'], 'dsz': md['dsz']})
                Constructor = eval(class_name)
                embeddings[key] = Constructor(key, **embed_args)

            model = cls.create(embeddings, **_state)
            model._state = _state
            model.load_weights(f"{basename}.wgt")

        return model

    @property
    def requires_state(self):
        pass


class AbstractGeneratorModel(LanguageModelBase):

    def create_layers(self, embeddings, **kwargs):
        self.embeddings = self.init_embed(embeddings, **kwargs)
        self.embeddings_proj = self.init_embeddings_proj(**kwargs)
        self.generator = self.init_generate(**kwargs)
        self.output_layer = self.init_output(embeddings, **kwargs)

    def call(self, inputs: Dict[str, TensorDef], hidden: TensorDef) -> Tuple[TensorDef, TensorDef]:
        emb = self.embed(inputs)
        output, hidden = self.generate(emb, hidden)
        return self.output_layer(output), hidden

    def embed(self, input):
        embedded_dropout = self.embeddings(input)
        return self.embeddings_proj(embedded_dropout)

    def init_embed(self, embeddings: Dict[str, TensorDef], **kwargs) -> BaseLayer:
        """This method creates the "embedding" layer of the inputs, with an optional reduction

        :param embeddings: A dictionary of embeddings

        :Keyword Arguments: See below
        * *embeddings_reduction* (defaults to `concat`) An operator to perform on a stack of embeddings
        * *embeddings_dropout = float(kwargs.get('embeddings_dropout', 0.0))

        :return: The output of the embedding stack followed by its reduction.  This will typically be an output
          with an additional dimension which is the hidden representation of the input
        """
        reduction = kwargs.get('embeddings_reduction', 'concat')
        embeddings_dropout = float(kwargs.get('embeddings_dropout', 0.0))
        return EmbeddingsStack({k: embeddings[k] for k in self.src_keys}, embeddings_dropout, reduction=reduction)

    def init_embeddings_proj(self, **kwargs):
        input_sz = self.embeddings.output_dim
        hsz = kwargs.get('hsz', kwargs.get('d_model'))
        if hsz != input_sz:
            proj = tf.keras.layers.Dense(hsz)
            print('Applying a transform from {} to {}'.format(input_sz, hsz))
        else:
            proj = PassThru(hsz)
        return proj

    def init_generate(self, **kwargs):
        pass

    def generate(self, emb, hidden):
        return self.generator((emb, hidden))

    def init_output(self, embeddings, **kwargs):
        self.vsz = embeddings[self.tgt_key].get_vsz()
        do_weight_tying = bool(kwargs.get('tie_weights', False))
        output_bias = kwargs.get('output_bias', False)
        if do_weight_tying:
            output = WeightTieDense(embeddings[self.tgt_key], use_bias=output_bias)
        else:
            output = tf.keras.layers.Dense(self.vsz)
        return output


@register_model(task='lm', name='default')
class RNNLanguageModel(AbstractGeneratorModel):
    """RNN-based Language Model built on base class
    """
    def __init__(self):
        """Construct an RNNLM
        """
        super().__init__()
        self.rnntype = 'lstm'
        self.initial_state = None

    @property
    def requires_state(self):
        return True

    def zero_state(self, inputs):
        batchsz = get_shape_as_list(inputs[self.src_keys[0]])[0]
        self.initial_state = self.generator.layer.zero_state(batchsz)

    def init_generate(self, **kwargs):
        """LSTM-based method for decoding

        :param inputs: The outputs of the embeddings
        :param kwargs: See above

        :return: The layer
        """
        pdrop = float(kwargs.get('dropout', 0.5))
        layers = kwargs.get('layers', kwargs.get('num_layers', 1))
        self.hsz = kwargs.get('hsz', kwargs.get('d_model'))
        return WithDropoutOnFirst(LSTMEncoderWithState(self.hsz, self.hsz, layers, pdrop, batch_first=True),
                                  pdrop,
                                  kwargs.get('variational', False))


@register_model(task='lm', name='transformer')
class TransformerLanguageModel(AbstractGeneratorModel):
    """Transformer-based Language Model built on base class
    """
    def __init__(self):
        """Construct an TLM
        """
        super().__init__()

    @property
    def requires_state(self):
        return False

    def init_generate(self, **kwargs):
        pdrop = float(kwargs.get('dropout', 0.1))
        layers = kwargs.get('layers', kwargs.get('num_layers', 1))
        d_model = int(kwargs.get('d_model', kwargs.get('hsz')))
        num_heads = kwargs.get('num_heads', 4)
        d_ff = int(kwargs.get('d_ff', 4 * d_model))
        rpr_k = kwargs.get('rpr_k')
        d_k = kwargs.get('d_k')
        scale = bool(kwargs.get('scale', True))
        activation = kwargs.get('activation', 'gelu')
        layer_norm_eps = kwargs.get('layer_norm_eps', 1e-12)
        layer_norms_after = kwargs.get('layer_norms_after', False)
        return TransformerEncoderStack(num_heads, d_model=d_model, pdrop=pdrop, scale=scale,
                                       layers=layers, d_ff=d_ff, rpr_k=rpr_k, d_k=d_k,
                                       activation=activation, layer_norm_eps=layer_norm_eps,
                                       layer_norms_after=layer_norms_after)

    def create_mask(self, bth):
        max_seqlen = get_shape_as_list(bth)[1]
        mask = subsequent_mask(max_seqlen)
        return mask

    def generate(self, bth, _):
        mask = self.create_mask(bth)
        return self.generator((bth, mask)), None


@register_model(task='lm', name='transformer-mlm')
class TransformerMaskedLanguageModel(TransformerLanguageModel):

    def create_mask(self, bth):
        nctx = get_shape_as_list(bth)[1]
        return tf.fill([1, 1, nctx, nctx], 1.)




import numpy as np
from baseline.remote import (
    RemoteModelREST,
    RemoteModelGRPC,
    register_remote,
    RemoteRESTClassifier,
    RemoteRESTTagger,
    RemoteRESTSeq2Seq,
    RemoteRESTEmbeddings,
    RemoteGRPCClassifier,
    RemoteGRPCTagger,
    RemoteGRPCSeq2Seq,
    RemoteGRPCEmbeddings,
)


class RemoteRESTTensorFlowMixin(RemoteModelREST):

    def create_request(self, examples):
        inputs = {}
        for feature in self.input_keys:
            tensor = examples[feature]
            if isinstance(tensor, np.ndarray):
                inputs[feature] = tensor.tolist()
            else:
                inputs[feature] = tensor
        request = {'signature_name': self.signature, 'inputs': inputs}
        return request

@register_remote('http-classify')
class RemoteRESTTensorFlowClassifier(RemoteRESTTensorFlowMixin, RemoteRESTClassifier): pass

@register_remote('http-tagger')
class RemoteRESTTensorFlowTagger(RemoteRESTTensorFlowMixin, RemoteRESTTagger): pass

@register_remote('http-seq2seq')
class RemoteRESTTensorFlowSeq2Seq(RemoteRESTTensorFlowMixin, RemoteRESTSeq2Seq): pass

@register_remote('http-servable-embeddings')
class RemoteRESTTensorFlowEmbeddings(RemoteRESTTensorFlowMixin, RemoteRESTEmbeddings): pass


@register_remote('grpc')
class RemoteGRPCTensorFlowMixin(RemoteModelGRPC): pass

@register_remote('grpc-classify')
class RemoteGRPCTensorFlowClassifier(RemoteGRPCTensorFlowMixin, RemoteGRPCClassifier): pass

@register_remote('grpc-tagger')
class RemoteGRPCTensorFlowTagger(RemoteGRPCTensorFlowMixin, RemoteGRPCTagger): pass

@register_remote('grpc-seq2seq')
class RemoteGRPCTensorFlowSeq2Seq(RemoteGRPCTensorFlowMixin, RemoteGRPCSeq2Seq): pass

@register_remote('grpc-servable-embeddings')
class RemoteGRPCTensorFlowEmbeddings(RemoteGRPCTensorFlowMixin, RemoteGRPCEmbeddings): pass


@register_remote('grpc-preproc')
class RemoteGRPCTensorFlowPreprocMixin(RemoteModelGRPC):

    def create_request(self, examples):
        # TODO: Remove TF dependency client side
        import tensorflow as tf

        request = self.predictpb.PredictRequest()
        request.model_spec.name = self.name
        request.model_spec.signature_name = self.signature
        if self.version is not None:
            request.model_spec.version.value = self.version

        for key in examples:
            if key.endswith('lengths'):
                shape = examples[key].shape
                tensor_proto = tf.contrib.util.make_tensor_proto(examples[key], shape=shape, dtype=tf.int32)
                request.inputs[key].CopyFrom(
                    tensor_proto
                )
            else:
                request.inputs[key].CopyFrom(
                    tf.contrib.util.make_tensor_proto(examples[key], shape=[len(examples[key]), 1])
                )
        return request

@register_remote('grpc-preproc-classify')
class RemoteGRPCTensorFlowPreprocClassifier(RemoteGRPCTensorFlowPreprocMixin, RemoteGRPCClassifier): pass

@register_remote('grpc-preproc-tagger')
class RemoteGRPCTensorFlowPreprocTagger(RemoteGRPCTensorFlowPreprocMixin, RemoteGRPCTagger): pass

@register_remote('grpc-preproc-seq2seq')
class RemoteGRPCTensorFlowPreprocSeq2Seq(RemoteGRPCTensorFlowPreprocMixin, RemoteGRPCSeq2Seq): pass

@register_remote('grpc-preproc-servable-embeddings')
class RemoteGRPCTensorFlowPreprocEmbeddings(RemoteGRPCTensorFlowPreprocMixin, RemoteGRPCEmbeddings): pass


@register_remote('http-preproc')
class RemoteRESTTensorFlowPreprocMixin(RemoteModelREST):

    def create_request(self, examples):
        inputs = {}
        if isinstance(examples['tokens'], np.ndarray):
            inputs['tokens'] = examples['tokens'].tolist()
        else:
            inputs['tokens'] = examples['tokens']
        for feature in self.input_keys:
            if feature.endswith('lengths'):
                if isinstance(examples[feature], np.ndarray):
                    inputs[feature] = examples[feature].tolist()
                else:
                    inputs[feature] = examples[feature]
        return {'signature_name': self.signature, 'inputs': inputs}

@register_remote('http-preproc-classify')
class RemoteRESTTensorFlowPreprocClassifier(RemoteRESTTensorFlowPreprocMixin, RemoteRESTClassifier): pass

@register_remote('http-preproc-tagger')
class RemoteRESTTensorFlowPreprocTagger(RemoteRESTTensorFlowPreprocMixin, RemoteRESTTagger): pass

@register_remote('http-preproc-seq2seq')
class RemoteRESTTensorFlowPreprocSeq2Seq(RemoteRESTTensorFlowPreprocMixin, RemoteRESTSeq2Seq): pass

@register_remote('http-preproc-servable-embeddings')
class RemoteRESTTensorFlowPreprocEmbeddings(RemoteRESTTensorFlowPreprocMixin, RemoteRESTEmbeddings): pass



from torch.utils.data.dataset import IterableDataset, TensorDataset
from baseline.vectorizers import Token1DVectorizer, BPEVectorizer1D
from baseline.progress import create_progress_bar
from baseline.reader import register_reader
from eight_mile.utils import str2bool, write_yaml, read_yaml, Offsets
import torch
import glob
import numpy as np
import os


class MultiFileLoader(IterableDataset):

    def __init__(self, directory, pattern, vocabs, vectorizers, nctx):
        super().__init__()
        self.src_vectorizer = vectorizers['src']
        self.tgt_vectorizer = vectorizers['tgt']
        self.pattern = pattern
        self.nctx = nctx
        self.directory = directory
        self.vocab = vocabs
        self.samples = 0
        self.rank = 0
        self.world_size = 1
        if torch.distributed.is_initialized():
            self.rank = torch.distributed.get_rank()
            self.world_size = torch.distributed.get_world_size()

        if os.path.exists(f"{directory}/md.yml"):
            f = read_yaml(f"{directory}/md.yml")
            self.samples = f['num_samples']
        else:
            files = list(glob.glob(f"{directory}/{self.pattern}"))
            pg = create_progress_bar(len(files))
            for file in pg(files):
                with open(file) as rf:
                    for _ in rf:
                        self.samples += 1
            write_yaml({'num_samples': self.samples}, f"{directory}/md.yml")

    def __len__(self):
        return self.samples

    def __iter__(self):
        # Each node has the same worker_info, so the unique offsets for each is
        # rank * num_workers + worker_id
        # and the total available workers is world_size * num_workers
        worker_info = torch.utils.data.get_worker_info()
        files = sorted(list(glob.glob(f"{self.directory}/{self.pattern}")))

        if worker_info is None:
            num_workers_per_node = 1
            node_worker_id = 0
        else:
            num_workers_per_node = worker_info.num_workers
            node_worker_id = worker_info.id
        all_workers = (self.world_size * num_workers_per_node)
        files_per_worker = len(files) // all_workers
        offset = self.rank * num_workers_per_node + node_worker_id
        start_idx = offset * files_per_worker
        end_idx = start_idx + files_per_worker if offset < all_workers - 1 else len(files)

        for file in files[start_idx:end_idx]:
            with open(file) as rf:
                for line in rf:
                    response = self.process_line(line)
                    if response:
                        yield response

    def process_line(self, line):
        """Read in a line and turn it into an entry

        The entries will get collated by the data loader

        :param line:
        :return:
        """

class Batcher(IterableDataset):

    def __init__(self, dataset, batchsz):
        self.batchsz = batchsz
        self.dataset = dataset

    def __len__(self):
        return len(self.dataset)//self.batchsz

    def _batch(self, batch_list):
        x = np.stack([item[0] for item in batch_list])
        y = np.stack([item[1] for item in batch_list])
        x_lens = np.stack([item[2] for item in batch_list])
        y_lens = np.stack([item[3] for item in batch_list])
        return {'src': x, 'tgt': y, 'src_lengths': x_lens, 'tgt_lengths': y_lens }

    def __iter__(self):

        dataset_iter = iter(self.dataset)
        steps_per_epoch = len(self.dataset)//self.batchsz
        for indices in range(steps_per_epoch):
            step = [next(dataset_iter) for _ in range(self.batchsz)]
            yield self._batch(step)


class NextTurnPredictionFileLoader(MultiFileLoader):

    def process_line(self, line):
        pair = line.strip().split('\t')
        # Unfortunately, this occassionally happens, a bunch of blank turns etc.
        if len(pair) != 2:
            return None
        q, r = pair
        q = q.strip()
        r = r.strip()
        if q == '' or r == '':
            return None
        q_vec, q_valid_lengths = self.src_vectorizer.run(reversed(q.split()), self.vocab)
        q_vec = np.roll(q_vec[::-1], -(self.src_vectorizer.mxlen - q_valid_lengths))
        r_vec, r_valid_lengths = self.tgt_vectorizer.run(['<GO>'] + r.split(), self.vocab)
        assert q_valid_lengths > 0 and q_valid_lengths <= self.src_vectorizer.mxlen
        assert r_valid_lengths > 0 and r_valid_lengths <= self.src_vectorizer.mxlen
        return q_vec, r_vec, q_valid_lengths, r_valid_lengths


@register_reader(task='seq2seq', name='paired-dir')
class MultiFileDatasetReader:
    """Provide a base-class to do operations that are independent of token representation
    """

    def __init__(self, vectorizers, trim, mxlen=64, pattern='*.txt', reader_type="ntp", **kwargs):
        self.nctx = mxlen
        self.pattern = pattern
        self.reader_type = reader_type
        self.vectorizers = vectorizers

    def build_vocabs(self, _=None, **kwargs):
        return {'src': self.vectorizers['src'].vocab}, self.vectorizers['tgt'].vocab

    def load(self, directory, src_vocabs, tgt_vocab, batch_size, **kwargs):
        # The src and target vectorizers are coupled
        loader = Batcher(NextTurnPredictionFileLoader(directory, self.pattern, tgt_vocab, self.vectorizers, self.nctx), batch_size)
        return loader




"""Train a classifier with TensorFlow

This module supports several different ways of training a model

1. feed_dict
2. datasets (`default` default for non-eager)
3. eager mode (`default` for eager)
4. distributed eager mode (`distributed`)
"""
from baseline.tf.classify.training import *



import os
from baseline.utils import exporter, optional_params

__all__ = []
export = exporter(__all__)


@export
class Exporter(object):

    def __init__(self, task, **kwargs):
        super(Exporter, self).__init__()
        self.task = task

    @classmethod
    def preproc_type(cls):
        return 'client'

    def _run(self, model_file, output_dir, project=None, name=None, model_version=None, **kwargs):
        raise NotImplementedError

    def run(self, model_file, output_dir, project=None, name=None, model_version=None, **kwargs):
        client_loc, server_loc = self._run(
            model_file,
            output_dir,
            project=project,
            name=name,
            model_version=model_version,
            **kwargs
        )
        if model_version is None:
            try:
                model_version = int(os.path.basename(client_loc))
            except ValueError:
                pass
        msg = {
            "client_bundle": client_loc,
            "server_bundle": server_loc,
            "project": project,
            "name": name,
            "version": model_version
        }
        for rep in self.task.reporting:
            rep.step(msg, 0, 'Export', 'EXPORT')
        self.task._close_reporting_hooks()
        return client_loc, server_loc


BASELINE_EXPORTERS = {}


@export
@optional_params
def register_exporter(cls, task, name=None):
    """Register an exporter

    Use this pattern if you want to provide an override to a `Exporter` class.

    """
    if name is None:
        name = cls.__name__

    if task not in BASELINE_EXPORTERS:
        BASELINE_EXPORTERS[task] = {}

    if name in BASELINE_EXPORTERS[task]:
        raise Exception('Error: attempt to re-defined previously registered handler {} in exporter registry'.format(name))

    BASELINE_EXPORTERS[task][name] = cls
    return cls


def create_exporter(task, name=None, **kwargs):
    return BASELINE_EXPORTERS[task.task_name()][name](task, **kwargs)



import os
import json
import baseline
from mead.exporters import register_exporter
from mead.preprocessors import create_preprocessors
from baseline.tf.embeddings import *
from baseline.utils import exporter, read_json
from collections import namedtuple
from mead.tf.exporters import ClassifyTensorFlowExporter, TaggerTensorFlowExporter, create_assets

__all__ = []
export = exporter(__all__)

SignatureOutput = namedtuple("SignatureOutput", ("classes", "scores"))


def get_string_size(tensor):
    """Get the number of tokens in a string sperated by space."""
    split = tf.string_split(tf.reshape(tensor, [-1]))
    return tf.shape(split)[1]


def get_mxlen(tensors):
    """Find the longest string in a batch."""
    sizes = tf.map_fn(
        get_string_size, tensors,
        dtype=tf.int32,
        back_prop=False
    )
    return tf.reduce_max(sizes)

def get_lengths(tensors):
    return tf.map_fn(
        get_string_size, tensors, dtype=tf.int32, back_prop=False
    )


def get_lengths(tensors):
    return tf.map_fn(
        get_string_size, tensors, dtype=tf.int32, back_prop=False
    )

def peek_lengths_key(model_file, field_map):
    """Check if there is lengths key to use when finding the length. (defaults to tokens)"""
    peek_state = read_json(model_file + ".state")
    lengths = peek_state.get('lengths_key', 'tokens')
    lengths = lengths.replace('_lengths', '')
    return field_map.get(lengths, lengths)


class PreProcessorController(object):
    def __init__(self, model_base_dir, pid, features, feature_exporter_field_map, length_field='tokens'):
        self.length_field = length_field
        saved_vectorizers = self.get_vectorizers(model_base_dir, pid)
        feature_names = [feature['name'] for feature in features]
        feature_vectorizer_mapping = {feature['name']: feature['vectorizer']['type'] for feature in features}
        self.preprocessors = dict()
        indices, vocabs = self.create_vocabs(model_base_dir, pid, feature_names)
        self.feature_exporter_field_map = feature_exporter_field_map
        self.FIELD_NAMES = list(set(self.feature_exporter_field_map.values()))
        for feature in feature_vectorizer_mapping:
            self.preprocessors[feature] = create_preprocessors(preprocessor_type=feature_vectorizer_mapping[feature],
                                                               feature=feature,
                                                               vectorizer=saved_vectorizers[feature],
                                                               index=indices[feature],
                                                               vocab=vocabs[feature])

    @staticmethod
    def get_vectorizers(model_base_dir, pid):
        """
        :model_file the path-like object to the model and model name.
        :vocab_suffixes the list of vocab types. e.g. 'word', 'char', 'ner'.
        """
        import pickle
        return pickle.load(open(os.path.join(model_base_dir, "vectorizers-{}.pkl".format(pid)), "rb"))

    def _create_example(self):
        serialized_tf_example = tf.compat.v1.placeholder(tf.string, name='tf_example')
        feature_configs = {f: tf.FixedLenFeature(shape=[], dtype=tf.string) for f in self.FIELD_NAMES}
        tf_example = tf.parse_example(serialized_tf_example, feature_configs)
        return tf_example

    def preproc(self, tf_example):
        preprocessed_inputs = {}
        for feature in self.preprocessors:
            preprocessed_inputs[feature] = self.preprocessors[feature].preproc(tf_example, self.mxlen)
        return preprocessed_inputs

    def create_preprocessed_input(self, tf_example):
        """
        Create a preprocessor chain inside of the tensorflow graph.
        """
        self.mxlen = get_mxlen(tf_example[self.length_field])

        types = {f: tf.int64 for f in self.preprocessors}
        return tf.map_fn(
            self.preproc, tf_example,
            dtype=types, back_prop=False
        )

    def create_vocabs(self, model_base_dir, pid, features):
        """
        :model_file the path-like object to the model and model name.
        :vocab_suffixes the list of vocab types. e.g. 'word', 'char', 'ner'.
        """
        indices = {}
        vocabs = {}
        for feature in features:
            feature_vocab_file = os.path.join(model_base_dir, "vocabs-{}-{}.json".format(feature, pid))
            if os.path.exists(feature_vocab_file):
                indices[feature], vocabs[feature] = self._read_vocab(feature_vocab_file, feature)
        return indices, vocabs

    @staticmethod
    def _read_vocab(vocab_file, feature_name):
        with open(vocab_file, 'r') as f:
            vocab = json.load(f)

        # Make a vocab list
        vocab_list = [''] * (len(vocab) + 1)

        for v, i in vocab.items():
            vocab_list[i] = v

        tok2index = tf.contrib.lookup.index_table_from_tensor(
            tf.constant(vocab_list),
            default_value=Offsets.UNK,
            dtype=tf.string,
            name='%s2index' % feature_name
        )
        return tok2index, vocab

    def run(self):
        tf_example = self._create_example()
        preprocessed = self.create_preprocessed_input(tf_example)
        lengths = get_lengths(tf_example[self.length_field])
        return tf_example, preprocessed, lengths


@export
@register_exporter(task='classify', name='preproc')
class ClassifyTensorFlowPreProcExporter(ClassifyTensorFlowExporter):

    @classmethod
    def preproc_type(cls):
        return 'server'

    def __init__(self, task, **kwargs):
        super(ClassifyTensorFlowPreProcExporter, self).__init__(task, **kwargs)
        self.feature_exporter_field_map = kwargs.get('feature_exporter_field_map', {'tokens': 'text'})
        self.return_labels = kwargs.get('return_labels', False)

    def _create_rpc_call(self, sess, model_file, **kwargs):
        model_base_dir = os.path.split(model_file)[0]
        pid = model_file.split("-")[-1]

        lengths = peek_lengths_key(model_file, self.feature_exporter_field_map)

        pc = PreProcessorController(model_base_dir, pid, self.task.config_params['features'],
                                    self.feature_exporter_field_map, lengths)
        tf_example, preprocessed, lengths = pc.run()
        # Create a dict of embedding names to sub-graph outputs to wire in as embedding inputs
        embedding_inputs = {}
        for feature in preprocessed:
            embedding_inputs[feature] = preprocessed[feature]
        model, classes, values = self._create_model(sess, model_file, lengths=lengths, **embedding_inputs)
        sig_input = {x: tf.saved_model.utils.build_tensor_info(tf_example[x]) for x in pc.FIELD_NAMES}
        if model.lengths is not None:
            sig_input.update({model.lengths_key: tf.saved_model.utils.build_tensor_info(model.lengths)})
        sig_output = SignatureOutput(classes, values)
        sig_name = 'predict_text'
        assets = create_assets(
            model_file,
            sig_input, sig_output, sig_name,
            model.lengths_key,
            return_labels=self.return_labels,
            preproc=self.preproc_type(),
        )
        return sig_input, sig_output, sig_name, assets


@export
@register_exporter(task='tagger', name='preproc')
class TaggerTensorFlowPreProcExporter(TaggerTensorFlowExporter):

    @classmethod
    def preproc_type(cls):
        return 'server'

    def __init__(self, task, **kwargs):
        super(TaggerTensorFlowPreProcExporter, self).__init__(task, **kwargs)
        self.feature_exporter_field_map = kwargs.get('feature_exporter_field_map', {'tokens': 'text'})
        self.return_labels = kwargs.get('return_labels', False)

    def _create_rpc_call(self, sess, model_file, **kwargs):
        model_base_dir = os.path.split(model_file)[0]
        pid = model_file.split("-")[-1]
        lengths = peek_lengths_key(model_file, self.feature_exporter_field_map)
        pc = PreProcessorController(model_base_dir, pid, self.task.config_params['features'],
                                    self.feature_exporter_field_map, lengths)
        tf_example, preprocessed, lengths = pc.run()
        # Create a dict of embedding names to sub-graph outputs to wire in as embedding inputs
        embedding_inputs = {}
        for feature in preprocessed:
            embedding_inputs[feature] = preprocessed[feature]
        model, classes, values = self._create_model(sess, model_file, lengths=lengths, **embedding_inputs)
        sig_input = {x: tf.saved_model.utils.build_tensor_info(tf_example[x]) for x in pc.FIELD_NAMES}
        sig_input.update({model.lengths_key: tf.saved_model.utils.build_tensor_info(model.lengths)})
        sig_output = SignatureOutput(classes, values)
        sig_name = 'tag_text'
        assets = create_assets(
            model_file,
            sig_input, sig_output, sig_name,
            model.lengths_key,
            return_labels=self.return_labels,
            preproc=self.preproc_type(),
        )
        return sig_input, sig_output, sig_name, assets



import numpy as np
from baseline.utils import listify
from baseline.vectorizers import Token1DVectorizer, register_vectorizer, _token_iterator


@register_vectorizer(name='text')
class Text1DVectorizer(Token1DVectorizer):
    def _next_element(self, tokens, vocab):
        for atom in self.iterable(tokens):
            yield atom

    def run(self, tokens, vocab):
        if self.mxlen < 0:
            self.mxlen = self.max_seen
        vec1d = np.full(self.mxlen, '', dtype=np.object)
        for i, atom in enumerate(self._next_element(tokens, vocab)):
            if i == self.mxlen:
                i -= 1
                break
            vec1d[i] = atom
        valid_length = i + 1

        if self.time_reverse:
            vec1d = vec1d[::-1]
            return vec1d, None
        return vec1d, valid_length


@register_vectorizer(name='dict_text')
class DictText1DVectorizer(Text1DVectorizer):
    def __init__(self, **kwargs):
        super(DictText1DVectorizer, self).__init__(**kwargs)
        self.fields = listify(kwargs.get('fields', 'text'))
        self.delim = kwargs.get('token_delim', '@@')

    def iterable(self, tokens):
        return _token_iterator(self, tokens)



import tensorflow as tf
from baseline.tf.tfy import TRAIN_FLAG
from eight_mile.utils import listify
from baseline.utils import get_model_file, get_metric_cmp
from baseline.train import create_trainer, register_training_func
from baseline.tf.seq2seq.training.utils import to_tensors, SHUF_BUF_SZ, NUM_PREFETCH


@register_training_func('seq2seq', 'dataset')
def fit_datasets(model_params, ts, vs, es=None, **kwargs):
    """
    Train an encoder-decoder network using TensorFlow with `tf.dataset`.  This
    is the default behavior for training.

    :param model_params: The model (or parameters to create the model) to train
    :param ts: A training data set
    :param vs: A validation data set
    :param es: A test data set, can be None
    :param kwargs:
        See below

    :Keyword Arguments:
        * *do_early_stopping* (``bool``) --
          Stop after evaluation data is no longer improving.  Defaults to True
        * *verbose* (`dict`) A dictionary containing `console` boolean and `file` name if on
        * *epochs* (``int``) -- how many epochs.  Default to 20
        * *outfile* -- Model output file, defaults to classifier-model.pyth
        * *patience* --
           How many epochs where evaluation is no longer improving before we give up
        * *reporting* --
           Callbacks which may be used on reporting updates
        * *nsteps* (`int`) -- If we should report every n-steps, this should be passed
        * *ema_decay* (`float`) -- If we are doing an exponential moving average, what decay to us4e
        * *clip* (`int`) -- If we are doing gradient clipping, what value to use
        * *optim* (`str`) -- The name of the optimizer we are using
        * *lr* (`float`) -- The learning rate we are using
        * *mom* (`float`) -- If we are using SGD, what value to use for momentum
        * *beta1* (`float`) -- Adam-specific hyper-param, defaults to `0.9`
        * *beta2* (`float`) -- Adam-specific hyper-param, defaults to `0.999`
        * *epsilon* (`float`) -- Adam-specific hyper-param, defaults to `1e-8

    :return: None
    """

    epochs = int(kwargs.get('epochs', 5))
    patience = int(kwargs.get('patience', epochs))
    model_file = get_model_file('seq2seq', 'tf', kwargs.get('basedir'))
    do_early_stopping = bool(kwargs.get('do_early_stopping', True))

    best_metric = 0
    if do_early_stopping:
        early_stopping_metric = kwargs.get('early_stopping_metric', 'perplexity')
        early_stopping_cmp, best_metric = get_metric_cmp(early_stopping_metric, kwargs.get('early_stopping_cmp'))
        patience = kwargs.get('patience', epochs)
        print('Doing early stopping on [%s] with patience [%d]' % (early_stopping_metric, patience))

    reporting_fns = listify(kwargs.get('reporting', []))
    print('reporting', reporting_fns)

    batchsz = kwargs['batchsz']
    ## First, make tf.datasets for ts, vs and es
    # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/distribute/README.md
    # effective_batch_sz = args.batchsz*args.gpus
    test_batchsz = kwargs.get('test_batchsz', batchsz)
    src_lengths_key = model_params.get('src_lengths_key')
    train_dataset = tf.data.Dataset.from_tensor_slices(to_tensors(ts, src_lengths_key))
    train_dataset = train_dataset.shuffle(buffer_size=SHUF_BUF_SZ)
    train_dataset = train_dataset.batch(batchsz, drop_remainder=False)
    train_dataset = train_dataset.repeat(epochs + 1)
    train_dataset = train_dataset.prefetch(NUM_PREFETCH)

    valid_dataset = tf.data.Dataset.from_tensor_slices(to_tensors(vs, src_lengths_key))
    valid_dataset = valid_dataset.batch(batchsz, drop_remainder=False)
    valid_dataset = valid_dataset.repeat(epochs + 1)
    valid_dataset = valid_dataset.prefetch(NUM_PREFETCH)

    iter = tf.compat.v1.data.Iterator.from_structure(tf.compat.v1.data.get_output_types(train_dataset),
                                                     tf.compat.v1.data.get_output_shapes(train_dataset))

    features, tgt = iter.get_next()
    # Add features to the model params
    model_params.update(features)
    # This is kind of crazy, but seems to work, hardwire a graph op for `mx_tgt_len`
    model_params.update({'tgt': tgt, 'mx_tgt_len': tf.reduce_max(features['tgt_len'])})

    # create the initialization operations
    train_init_op = iter.make_initializer(train_dataset)
    valid_init_op = iter.make_initializer(valid_dataset)

    TRAIN_FLAG()
    trainer = create_trainer(model_params, **kwargs)

    last_improved = 0

    for epoch in range(epochs):
        trainer.sess.run(train_init_op)
        trainer.train(ts, reporting_fns)
        trainer.sess.run(valid_init_op)
        test_metrics = trainer.test(vs, reporting_fns, phase='Valid')

        if do_early_stopping is False:
            trainer.checkpoint()
            trainer.model.save(model_file)

        elif early_stopping_cmp(test_metrics[early_stopping_metric], best_metric):
            last_improved = epoch
            best_metric = test_metrics[early_stopping_metric]
            print('New best %.3f' % best_metric)
            trainer.checkpoint()
            trainer.model.save(model_file)

        elif (epoch - last_improved) > patience:
            print('Stopping due to persistent failures to improve')
            break

    if do_early_stopping is True:
        print('Best performance on %s: %.3f at epoch %d' % (early_stopping_metric, best_metric, last_improved))

    if es is not None:
        print('Reloading best checkpoint')
        trainer.recover_last_checkpoint()

        test_dataset = tf.data.Dataset.from_tensor_slices(to_tensors(es, src_lengths_key))
        test_dataset = test_dataset.batch(test_batchsz, drop_remainder=False)
        test_dataset = test_dataset.repeat(epochs + 1)
        test_dataset = test_dataset.prefetch(NUM_PREFETCH)
        test_init_op = iter.make_initializer(test_dataset)

        trainer.sess.run(test_init_op)
        trainer.test(es, reporting_fns, phase='Test')



import os
import shutil
import logging
import datetime
from collections import namedtuple
import baseline
from baseline.tf.embeddings import *
from baseline.utils import (
    exporter,
    Offsets,
    ls_props,
    read_json,
    write_json,
    transition_mask
)
from baseline.model import load_tagger_model, load_model, load_seq2seq_model
from mead.exporters import Exporter, register_exporter
from mead.utils import save_to_bundle, create_metadata, get_output_paths


__all__ = []
export = exporter(__all__)


FIELD_NAME = 'text/tokens'
ASSET_FILE_NAME = 'model.assets'
logger = logging.getLogger('mead')
SignatureOutput = namedtuple("SignatureOutput", ("classes", "scores"))


def get_tf_index_from_unzipped(dir_path):
    return os.path.join(dir_path, [x[:-6] for x in os.listdir(dir_path) if 'index' in x][0])


@export
class TensorFlowExporter(Exporter):

    def __init__(self, task, **kwargs):
        super(TensorFlowExporter, self).__init__(task, **kwargs)

    def _run(self, basename, output_dir, project=None, name=None, model_version=None, **kwargs):
        basename = get_tf_index_from_unzipped(basename)

        with tf.compat.v1.Graph().as_default():
            config_proto = tf.compat.v1.ConfigProto(allow_soft_placement=True)
            with tf.compat.v1.Session(config=config_proto) as sess:
                sig_input, sig_output, sig_name, assets = self._create_rpc_call(sess, basename)
                client_output, server_output = get_output_paths(
                    output_dir,
                    project, name,
                    model_version,
                    kwargs.get('remote', True), make_server=False
                )
                logger.info('Saving vectorizers and vocabs to %s', client_output)
                logger.info('Saving serialized model to %s' % server_output)
                try:
                    builder = self._create_saved_model_builder(sess, server_output, sig_input, sig_output, sig_name)
                    create_bundle(builder, client_output, basename, assets)
                    logger.info('Successfully exported model to %s' % output_dir)
                except AssertionError as e:
                    # model already exists
                    raise e
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    # export process broke.
                    # TODO(MB): we should remove the directory, if one has been saved already.
                    raise e
        return client_output, server_output

    def _create_saved_model_builder(self, sess, output_path, sig_input, sig_output, sig_name):
        """
        create the SavedModelBuilder with standard endpoints.

        we reuse the classify constants from tensorflow to define the predict
        endpoint so that we can call the output by classes/scores.
        """
        builder = tf.compat.v1.saved_model.builder.SavedModelBuilder(output_path)

        classes_output_tensor = tf.compat.v1.saved_model.utils.build_tensor_info(
            sig_output.classes)

        output_def_map = {
            tf.compat.v1.saved_model.signature_constants.CLASSIFY_OUTPUT_CLASSES: classes_output_tensor
        }
        if sig_output.scores is not None:
            scores_output_tensor = tf.compat.v1.saved_model.utils.build_tensor_info(sig_output.scores)
            output_def_map[tf.compat.v1.saved_model.signature_constants.CLASSIFY_OUTPUT_SCORES] = scores_output_tensor

        prediction_signature = (
            tf.compat.v1.saved_model.signature_def_utils.build_signature_def(
                inputs=sig_input,
                outputs=output_def_map,  # we reuse classify constants here.
                method_name=tf.compat.v1.saved_model.signature_constants.PREDICT_METHOD_NAME
            )
        )

        legacy_init_op = tf.group(tf.compat.v1.tables_initializer(), name='legacy_init_op')
        definition = dict({})
        definition[sig_name] = prediction_signature
        builder.add_meta_graph_and_variables(
            sess, [tf.compat.v1.saved_model.tag_constants.SERVING],
            signature_def_map=definition,
            legacy_init_op=legacy_init_op)

        return builder

    def _create_rpc_call(self, sess, basename, **kwargs):
        pass


@export
@register_exporter(task='classify', name='default')
class ClassifyTensorFlowExporter(TensorFlowExporter):

    def __init__(self, task, **kwargs):
        super(ClassifyTensorFlowExporter, self).__init__(task, **kwargs)
        self.return_labels = kwargs.get('return_labels', True)

    def _create_model(self, sess, basename, **kwargs):
        model = load_model(basename, sess=sess, **kwargs)
        values, indices = tf.nn.top_k(model.probs, len(model.labels))
        # Restore the checkpoint
        if self.return_labels:
            class_tensor = tf.constant(model.labels)
            table = tf.contrib.lookup.index_to_string_table_from_tensor(class_tensor)
            classes = table.lookup(tf.to_int64(indices))
            return model, classes, values
        else:
            return model, indices, values

    def _create_rpc_call(self, sess, basename, **kwargs):
        model, classes, values = self._create_model(sess, basename)

        predict_tensors = {}
        predict_tensors[model.lengths_key] = tf.compat.v1.saved_model.utils.build_tensor_info(model.lengths)

        for k, v in model.embeddings.items():
            try:
                predict_tensors[k] = tf.compat.v1.saved_model.utils.build_tensor_info(v.x)
            except:
                raise Exception('Unknown attribute in signature: {}'.format(v))

        sig_input = predict_tensors
        sig_output = SignatureOutput(classes, values)
        sig_name = 'predict_text'

        assets = create_assets(
            basename,
            sig_input, sig_output, sig_name,
            model.lengths_key,
            return_labels=self.return_labels,
            preproc=self.preproc_type()
        )
        return sig_input, sig_output, sig_name, assets


@export
@register_exporter(task='tagger', name='default')
class TaggerTensorFlowExporter(TensorFlowExporter):

    def __init__(self, task, **kwargs):
        super(TaggerTensorFlowExporter, self).__init__(task, **kwargs)
        self.return_labels = kwargs.get('return_labels', False)  # keep default behavior

    def _create_model(self, sess, basename, **kwargs):
        model = load_tagger_model(basename, sess=sess, **kwargs)
        softmax_output = tf.nn.softmax(model.probs)
        values, _ = tf.nn.top_k(softmax_output, 1)
        indices = model.best
        if self.return_labels:
            labels = read_json(basename + '.labels')
            list_of_labels = [''] * len(labels)
            for label, idval in labels.items():
                list_of_labels[idval] = label
            class_tensor = tf.constant(list_of_labels)
            table = tf.contrib.lookup.index_to_string_table_from_tensor(class_tensor)
            classes = table.lookup(tf.to_int64(indices))
            return model, classes, values
        else:
            return model, indices, values

    def _create_rpc_call(self, sess, basename, **kwargs):
        model, classes, values = self._create_model(sess, basename)

        predict_tensors = {}
        predict_tensors[model.lengths_key] = tf.compat.v1.saved_model.utils.build_tensor_info(model.lengths)

        for k, v in model.embeddings.items():
            try:
                predict_tensors[k] = tf.compat.v1.saved_model.utils.build_tensor_info(v.x)
            except:
                raise Exception('Unknown attribute in signature: {}'.format(v))

        sig_input = predict_tensors
        sig_output = SignatureOutput(classes, values)
        sig_name = 'tag_text'
        assets = create_assets(
            basename,
            sig_input, sig_output, sig_name,
            model.lengths_key,
            return_labels=self.return_labels,
            preproc=self.preproc_type()
        )
        return sig_input, sig_output, sig_name, assets


@export
@register_exporter(task='seq2seq', name='default')
class Seq2SeqTensorFlowExporter(TensorFlowExporter):
    """
    seq2seq has a source and tgt embedding layer.

    Please note that Tensorflow creates a state file where
    the target embeddings are not a dictionary. In order to reuse
    the embedding creation logic, I initialize a list of tuples
    as if items() was called on a dictionary. This also hides
    the correct key for target embeddings, so that is stored
    as a constant alongside the state keys.

    Perhaps this should be modified in the model, but this would decrease
    the readablity of the state file. It also doesn't make sense
    to generalize target embeddings to a dict since it's doubtful
    that we will ever target more than one embedding.

    :see baseline.python.tf.seq2seq.model:L411
    """
    SOURCE_STATE_EMBED_KEY = 'src_embeddings'
    TARGET_STATE_EMBED_KEY = 'tgt_embedding'
    TARGET_EMBED_KEY = 'tgt'

    def __init__(self, task, **kwargs):
        super(Seq2SeqTensorFlowExporter, self).__init__(task, **kwargs)
        self.return_labels = kwargs.get('return_labels', False)

    def _create_model(self, sess, basename, **kwargs):
        model = load_seq2seq_model(
            basename,
            sess=sess, predict=True,
            beam=self.task.config_params.get('beam', 30),
            **kwargs
        )
        return model, model.decoder.best, None

    def _create_rpc_call(self, sess, basename):
        model, classes, values = self._create_model(sess, basename)

        predict_tensors = {}
        predict_tensors[model.src_lengths_key] = tf.compat.v1.saved_model.utils.build_tensor_info(model.src_len)

        for k, v in model.src_embeddings.items():
            try:
                predict_tensors[k] = tf.compat.v1.saved_model.utils.build_tensor_info(v.x)
            except:
                raise Exception('Unknown attribute in signature: {}'.format(v))

        sig_input = predict_tensors
        sig_output = SignatureOutput(classes, values)
        sig_name = 'suggest_text'
        assets = create_assets(
            basename,
            sig_input, sig_output, sig_name,
            model.src_lengths_key,
            beam=model.decoder.beam_width,
            return_labels=self.return_labels,
            preproc=self.preproc_type(),
        )

        return sig_input, sig_output, sig_name, assets


def create_bundle(builder, output_path, basename, assets=None):
    """Creates the output files for an exported model.

    :builder the tensorflow saved_model builder.
    :output_path the path to export a model. this includes the model_version name.
    :assets a dictionary of assets to save alongside the model.
    """
    builder.save()

    model_name = os.path.basename(basename)
    directory = os.path.realpath(os.path.dirname(basename))
    save_to_bundle(output_path, directory, assets)


def create_assets(basename, sig_input, sig_output, sig_name, lengths_key=None, beam=None, return_labels=False, preproc='client'):
    """Save required variables for running an exported model from baseline's services.

    :basename the base model name. e.g. /path/to/tagger-26075
    :sig_input the input dictionary
    :sig_output the output namedTuple
    :lengths_key the lengths_key from the model.
        used to translate batch output. Exported models will return a flat list,
        and it needs to be reshaped into per-example lists. We use this key to tell
        us which feature holds the sequence lengths.

    """
    inputs = [k for k in sig_input]
    outputs =  sig_output._fields
    model_name = basename.split("/")[-1]
    directory = basename.split("/")[:-1]
    metadata = create_metadata(inputs, outputs, sig_name, model_name, lengths_key, beam=beam,
                               return_labels=return_labels, preproc=preproc)
    return metadata



import logging
from baseline.pytorch.torchy import *
from baseline.utils import Offsets, write_json
from baseline.model import TaggerModel
from baseline.model import register_model
import torch.autograd
import os

logger = logging.getLogger('baseline')


class TaggerModelBase(nn.Module, TaggerModel):
    """Base class for tagger models

    This class provides the model base for tagging.  To create a tagger, overload `create_layers()` and `forward()`.
    Most implementations should be able to subclass the `AbstractEncoderTaggerModel`, which inherits from this and imposes
    additional structure
    """
    def __init__(self):
        """Constructor"""
        super().__init__()
        self.gpu = False

    def save(self, outname: str):
        """Save out the model

        :param outname: The name of the checkpoint to write
        :return:
        """
        torch.save(self, outname)
        basename, _ = os.path.splitext(outname)
        write_json(self.labels, basename + ".labels")

    def cuda(self, device=None):
        self.gpu = True
        return super().cuda(device=device)


    @staticmethod
    def load(filename: str, **kwargs) -> 'TaggerModelBase':
        """Create and load a tagger model from file

        """
        device = kwargs.get('device')
        if not os.path.exists(filename):
            filename += '.pyt'
        model = torch.load(filename, map_location=device)
        model.gpu = False if device == 'cpu' else model.gpu
        return model

    def drop_inputs(self, key, x):
        """Do dropout on inputs, using the dropout value (or none if not set)
        This works by applying a dropout mask with the probability given by a
        value within the `dropin_values: Dict[str, float]`, keyed off the text name
        of the feature

        :param key: The feature name
        :param x: The tensor to drop inputs for
        :return: The dropped out tensor
        """
        v = self.dropin_values.get(key, 0)
        if not self.training or v == 0:
            return x

        mask_pad = x != Offsets.PAD
        mask_drop = x.new(x.size(0), x.size(1)).bernoulli_(v).to(mask_pad.dtype)
        x.masked_fill_(mask_pad & mask_drop, Offsets.UNK)
        return x

    def input_tensor(self, key, batch_dict, perm_idx, numpy_to_tensor=False):
        """Given a batch of input, and a key, prepare and noise the input

        :param key: The key of the tensor
        :param batch_dict: The batch of data as a dictionary of tensors
        :param perm_idx: The proper permutation order to get lengths descending
        :param numpy_to_tensor: Should this tensor be converted to a `torch.Tensor`
        :return:
        """
        tensor = batch_dict[key]
        if numpy_to_tensor:
            tensor = torch.from_numpy(tensor)

        tensor = self.drop_inputs(key, tensor)
        tensor = tensor[perm_idx]
        if self.gpu:
            tensor = tensor.cuda()
        return tensor

    def make_input(self, batch_dict: Dict[str, TensorDef], perm: bool = False, numpy_to_tensor: bool = False) -> Dict[str, TensorDef]:
        """Transform a `batch_dict` into format suitable for tagging

        :param batch_dict: A dictionary containing all inputs to the embeddings for this model
        :param perm: Should we sort data by length descending?
        :param numpy_to_tensor: Do we need to convert the input from numpy to a torch.Tensor?
        :return: A dictionary representation of this batch suitable for processing
        """
        example_dict = dict({})
        lengths = batch_dict[self.lengths_key]
        if numpy_to_tensor:
            lengths = torch.from_numpy(lengths)

        lengths, perm_idx = lengths.sort(0, descending=True)
        if self.gpu:
            lengths = lengths.cuda()

        example_dict['lengths'] = lengths
        for key in self.embeddings.keys():
            example_dict[key] = self.input_tensor(key, batch_dict, perm_idx, numpy_to_tensor=numpy_to_tensor)
        y = batch_dict.get('y')
        if y is not None:
            if numpy_to_tensor:
                y = torch.from_numpy(y)
            y = y[perm_idx]
            if self.gpu:
                y = y.cuda()
            example_dict['y'] = y

        ids = batch_dict.get('ids')
        if ids is not None:
            if numpy_to_tensor:
                ids = torch.from_numpy(ids)
            ids = ids[perm_idx]
            if self.gpu:
                ids = ids.cuda()
            example_dict['ids'] = ids
        if perm:
            return example_dict, perm_idx
        return example_dict

    def get_labels(self) -> List[str]:
        """Get the labels (names of each class)

        :return: (`List[str]`) The labels
        """
        return self.labels

    def predict(self, batch_dict: Dict[str, TensorDef], **kwargs) -> TensorDef:
        """Take in a batch of data, and predict the tags

        :param batch_dict: A batch of features that is to be predicted
        :param kwargs: See Below

        :Keyword Arguments:

        * *numpy_to_tensor* (``bool``) Should we convert input from numpy to `torch.Tensor` Defaults to `True`
        :return: A batch-sized tensor of predictions
        """
        numpy_to_tensor = bool(kwargs.get('numpy_to_tensor', True))
        inputs, perm_idx = self.make_input(batch_dict, perm=True, numpy_to_tensor=numpy_to_tensor)
        outputs = self(inputs)
        return unsort_batch(outputs, perm_idx)

    @classmethod
    def create(cls, embeddings: Dict[str, TensorDef], labels: List[str], **kwargs) -> 'TaggerModelBase':
        """Create a tagger from the inputs.  Most classes shouldnt extend this

        :param embeddings: A dictionary containing the input feature indices
        :param labels: A list of the labels (tags)
        :param kwargs: See below

        :Keyword Arguments:

        * *lengths_key* (`str`) Which feature identifies the length of the sequence
        * *activation* (`str`) What type of activation function to use (defaults to `tanh`)
        * *dropout* (`str`) What fraction dropout to apply
        * *dropin* (`str`) A dictionarwith feature keys telling what fraction of word masking to apply to each feature

        :return:
        """
        model = cls()
        model.lengths_key = kwargs.get('lengths_key')
        model.activation_type = kwargs.get('activation', 'tanh')
        model.pdrop = float(kwargs.get('dropout', 0.5))
        model.dropin_values = kwargs.get('dropin', {})
        model.labels = labels
        model.gpu = not bool(kwargs.get('nogpu', False))
        model.create_layers(embeddings, **kwargs)
        return model

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        """This method defines the model itself, and must be overloaded by derived classes

        This function will update `self` with the layers required to execute the `call()` method

        :param embeddings: The input feature indices
        :param kwargs:
        :return:
        """
    def compute_loss(self, inputs):
        """Define a loss function from the inputs, which includes the gold tag values as `inputs['y']`

        :param inputs:
        :return:
        """


class AbstractEncoderTaggerModel(TaggerModelBase):
    """Class defining a typical flow for taggers.  Most taggers should extend this class

    This class provides the model base for tagging by providing specific hooks for each phase.  There are
    4 basic steps identified in this class:

    1. embed
    2. encode (transduction)
    3. proj (projection to the final number of labels)
    4. decode

    There is an `init_* method for each of this phases, allowing you to
    define and return a custom layer.

    The actual forward method is defined as a combination of these 3 steps, which includes a
    projection from the encoder output to the number of labels.

    Decoding in taggers refers to the process of selecting the best path through the labels and is typically
    implemented either as a constrained greedy decoder or as a CRF layer
    """
    def __init__(self):
        """Constructor"""
        super().__init__()

    def init_embed(self, embeddings: Dict[str, TensorDef], **kwargs) -> BaseLayer:
        """This method creates the "embedding" layer of the inputs, with an optional reduction

        :param embeddings: A dictionary of embeddings

        :Keyword Arguments: See below
        * *embeddings_reduction* (defaults to `concat`) An operator to perform on a stack of embeddings

        :return: The output of the embedding stack followed by its reduction.  This will typically be an output
          with an additional dimension which is the hidden representation of the input
        """
        return EmbeddingsStack(embeddings, self.pdrop, reduction=kwargs.get('embeddings_reduction', 'concat'))

    def init_encode(self, input_dim, **kwargs) -> BaseLayer:
        """Provide a layer object that represents the `encode` phase of the model
        :param input_dim: The hidden input size
        :param kwargs:
        :return: The encoder
        """

    def init_proj(self, **kwargs) -> BaseLayer:
        """Provide a projection from the encoder output to the number of labels

        This projection typically will not include any activation, since its output is the logits that
        the decoder is built on

        :param kwargs:
        :return: A projection from the encoder output size to the final number of labels
        """
        return Dense(self.encoder.output_dim, len(self.labels))

    def init_decode(self, **kwargs) -> BaseLayer:
        """Define a decoder from the inputs

        :param kwargs: See below
        :keyword Arguments:
        * *crf* (``bool``) Should we create a CRF as the decoder
        * *constraint_mask* (``tensor``) A constraint mask to apply to the transitions
        * *reduction* (``str``) How to reduce the loss, defaults to `batch`
        :return: A decoder layer
        """
        use_crf = bool(kwargs.get('crf', False))
        constraint_mask = kwargs.get('constraint_mask')
        if constraint_mask is not None:
            constraint_mask = constraint_mask.unsqueeze(0)

        if use_crf:
            decoder = CRF(len(self.labels), constraint_mask=constraint_mask, batch_first=True)
        else:
            decoder = TaggerGreedyDecoder(
                len(self.labels),
                constraint_mask=constraint_mask,
                batch_first=True,
                reduction=kwargs.get('reduction', 'batch')
            )
        return decoder

    def create_layers(self, embeddings: Dict[str, TensorDef], **kwargs):
        """This class overrides this method to produce the outline of steps for a transduction tagger

        :param embeddings: The input embeddings dict
        :param kwargs:
        :return:
        """
        self.embeddings = self.init_embed(embeddings, **kwargs)
        self.encoder = self.init_encode(self.embeddings.output_dim, **kwargs)
        self.proj_layer = self.init_proj(**kwargs)
        self.decoder = self.init_decode(**kwargs)

    def transduce(self, inputs: Dict[str, TensorDef]) -> TensorDef:
        """This operation performs embedding of the input, followed by encoding and projection to logits

        :param inputs: The feature indices to embed
        :return: Transduced (post-encoding) output
        """
        lengths = inputs["lengths"]
        embedded = self.embeddings(inputs)
        embedded = (embedded, lengths)
        transduced = self.proj_layer(self.encoder(embedded))
        return transduced

    def decode(self, tensor: TensorDef, lengths: TensorDef) -> TensorDef:
        """Take in the transduced (encoded) input and decode it

        :param tensor: Transduced input
        :param lengths: Valid lengths of the transduced input
        :return: A best path through the output
        """
        return self.decoder((tensor, lengths))

    def forward(self, inputs: Dict[str, TensorDef]) -> TensorDef:
        """Take the input and produce the best path of labels out

        :param inputs: The feature indices for the input
        :return: The most likely path through the output labels
        """
        transduced = self.transduce(inputs)
        path = self.decode(transduced, inputs.get("lengths"))
        return path

    def compute_loss(self, inputs):
        """Provide the loss by requesting it from the decoder

        :param inputs: A batch of inputs
        :return:
        """
        tags = inputs['y']
        lengths = inputs['lengths']
        unaries = self.transduce(inputs)
        return self.decoder.neg_log_loss(unaries, tags, lengths)


@register_model(task='tagger', name='default')
class RNNTaggerModel(AbstractEncoderTaggerModel):
    """RNN-based tagger implementation: this is the default tagger for mead-baseline

    Overload the encoder, typically as a BiLSTM
    """
    def __init__(self):
        super().__init__()

    def init_encode(self, input_dim: int, **kwargs) -> BaseLayer:
        """Override the base method to produce an RNN transducer

        :param input_dim: The size of the input
        :param kwargs: See below

        :Keyword Arguments:
        * *rnntype* (``str``) The type of RNN, defaults to `blstm`
        * *layers* (``int``) The number of layers to stack
        * *hsz* (``int``) The number of hidden units for each layer in the encoder
        * *dropout* (``float``) The dropout rate
        * *weight_init* (``str``) The weight initializer, defaults to `uniform`
        * *unif* (``float``) A value for the weight initializer
        :return: An encoder
        """
        rnntype = kwargs.get('rnntype', 'blstm')
        nlayers = int(kwargs.get('layers', 1))
        unif = kwargs.get('unif', 0)
        hsz = int(kwargs['hsz'])
        pdrop = float(kwargs.get('dropout', 0.5))
        weight_init = kwargs.get('weight_init', 'uniform')
        Encoder = LSTMEncoderSequence if rnntype == 'lstm' else BiLSTMEncoderSequence
        return Encoder(input_dim, hsz, nlayers, pdrop, unif=unif, initializer=weight_init, batch_first=True)


@register_model(task='tagger', name='transformer')
class TransformerTaggerModel(AbstractEncoderTaggerModel):
    """Transformer-based tagger model

    Overload the encoder using a length-aware Transformer
    """
    def __init__(self):
        super().__init__()

    def init_encode(self, input_dim: int, **kwargs) -> BaseLayer:
        """Override the base method to produce an RNN transducer

        :param input_dim: The size of the input
        :param kwargs: See below

        :Keyword Arguments:
        * *num_heads* (``int``) The number of heads for multi-headed attention
        * *layers* (``int``) The number of layers to stack
        * *hsz* (``int``) The number of hidden units for each layer in the encoder
        * *dropout* (``float``) The dropout rate, defaults
        * *d_ff* (``int``) The feed-forward layer size
        * *rpr_k* (``list`` or ``int``) The relative attention sizes.  If its a list, one scalar per layer, if its
          a scalar, apply same size to each layer
        :return: An encoder
        """
        layers = int(kwargs.get('layers', 1))
        num_heads = int(kwargs.get('num_heads', 4))
        pdrop = float(kwargs.get('dropout', 0.5))
        scale = False
        hsz = int(kwargs['hsz'])
        rpr_k = kwargs.get('rpr_k', 100)
        d_ff = kwargs.get('d_ff')
        encoder = TransformerEncoderStackWithLengths(num_heads, hsz, pdrop, scale, layers, d_ff=d_ff, rpr_k=rpr_k, input_sz=input_dim)
        return encoder


@register_model(task='tagger', name='cnn')
class CNNTaggerModel(AbstractEncoderTaggerModel):
    """Convolutional (AKA TDNN) tagger

    Overload the encoder using a conv layer

    """
    def __init__(self):
        super().__init__()

    def init_encode(self, input_dim: int, **kwargs) -> BaseLayer:
        """Override the base method to produce an RNN transducer

        :param input_dim: The size of the input
        :param kwargs: See below

        :Keyword Arguments:
        * *layers* (``int``) The number of layers to stack
        * *hsz* (``int``) The number of hidden units for each layer in the encoder
        * *dropout* (``float``) The dropout rate, defaults
        * *activation_type* (``str``) Defaults to `relu`
        * *wfiltsz* (``int``) The 1D filter size for the convolution
        :return: An encoder
        """
        layers = int(kwargs.get('layers', 1))
        pdrop = float(kwargs.get('dropout', 0.5))
        filtsz = kwargs.get('wfiltsz', 5)
        activation_type = kwargs.get('activation_type', 'relu')
        hsz = int(kwargs['hsz'])
        return WithoutLength(ConvEncoderStack(input_dim, hsz, filtsz, layers, pdrop, activation_type))


@register_model(task='tagger', name='pass')
class PassThruTaggerModel(AbstractEncoderTaggerModel):
    """A Pass-thru implementation of the encoder

    When we fine-tune our taggers from things like BERT embeddings, we might want to just pass through our
    embedding result directly to the output decoder.  This model provides a mechanism for this by providing
    a simple identity layer
    """
    def __init__(self):
        super().__init__()

    def init_encode(self, input_dim: int, **kwargs) -> BaseLayer:
        """Identity layer encoder

        :param input_dim: The input dims
        :param kwargs: None
        :return: An encoder
        """
        return WithoutLength(PassThru(input_dim))



import time
import logging
import numpy as np
import eight_mile.optz
from eight_mile.optz import *
from eight_mile.utils import exporter, optional_params
from baseline.utils import register
import math


__all__ = []
__all__.extend(eight_mile.optz.__all__)
export = exporter(__all__)


@export
class Trainer:

    def __init__(self):
        self.train_epochs = 0
        self.valid_epochs = 0
        self.nsteps = None
        self.nstep_agg = 0
        self.nstep_div = 0
        self.nstep_start = 0
        self.log = logging.getLogger('baseline.timing')

    def report(self, step, metrics, start, phase, tt, reporting_fns, steps=1):
        """Make a report (both metric and timinging).

        :param step: `int` The step number of this report (epoch or nstep number).
        :param metrics: `dict` The metrics to report.
        :param start: `int` The starting time of this segment.
        :param phase: `str` The phase type. {'Train', 'Valid', 'Test'}
        :param tt: `str` The tick type. {'STEP', 'EPOCH'}
        :param reporting_fns: `List[Callable]` The list of reporting functions to call.
        :param steps: `int` The number of steps in this segment, used to normalize the time.
        """
        elapsed = time.time() - start
        for reporting in reporting_fns:
            reporting(metrics, step, phase, tt)
        self.log.debug({
            'tick_type': tt, 'tick': step, 'phase': phase,
            'time': elapsed / float(steps),
            'step/sec': steps / float(elapsed)
        })

    def reset_nstep(self):
        self.nstep_agg = 0
        self.nstep_div = 0
        self.nstep_start = time.time()

    @staticmethod
    def calc_metrics(agg, norm):
        return {'avg_loss': agg / float(norm)}

    def test(self, loader, reporting_fns):
        pass

    def train(self, loader, reporting_fns):
        pass


@export
class EpochReportingTrainer(Trainer):

    def __init__(self):
        super().__init__()

    def train(self, ts, reporting_fns, **kwargs):
        start_time = time.time()
        self.nstep_start = start_time
        metrics = self._train(ts, reporting_fns=reporting_fns, **kwargs)
        self.train_epochs += 1
        self.report(
            self.train_epochs, metrics, start_time,
            'Train', 'EPOCH', reporting_fns
        )
        return metrics

    def test(self, vs, reporting_fns, phase='Valid', **kwargs):
        start_time = time.time()
        metrics = self._test(vs, **kwargs)
        epochs = 0
        if phase == 'Valid':
            self.valid_epochs += 1
            epochs = self.valid_epochs
        self.report(
            epochs, metrics, start_time,
            phase, 'EPOCH', reporting_fns
        )
        return metrics

    def _train(self, ts):
        pass

    def _test(self, vs, **kwargs):
        pass


BASELINE_TRAINERS = {}


@export
@optional_params
def register_trainer(cls, task, name=None):
    """Register a function as a plug-in

    Use this pattern if you want to provide an override to a `Trainer` class.

    """
    if task not in BASELINE_TRAINERS:
        BASELINE_TRAINERS[task] = {}
    if name is None:
        name = cls.__name__
    BASELINE_TRAINERS[task][name] = cls
    return cls


BASELINE_FIT_FUNC = {}


@export
@optional_params
def register_training_func(func, task, name=None):
    """Register a training by-pass

    Use this pattern if you want to change the entire training hook.  Your function will have to fulfill all the
    behaviors that fit() normally handles

    :param func:
    :param name:
    :return:
    """
    if task not in BASELINE_FIT_FUNC:
        BASELINE_FIT_FUNC[task] = {}
    if name is None:
        name = 'default'

    if name in BASELINE_FIT_FUNC[task]:
        raise Exception('Attempting to override the fit function without providing a suitable name')

    BASELINE_FIT_FUNC[task][name] = func
    return func


@export
def fit(model_params, ts, vs, es, **kwargs):
    """This method delegates to the registered fit function for each DL framework.  It is possible to provide a by-pass
    to our defined fit functions for each method (this is considered advanced usage).  In cases where the user wishes
    to provide their own fit hook, the need to decorate the bypass hook with @register_training_func(name='myname'),
    and then pass in the `fit_func='myname'` to this.  MEAD handles this automatically -- just pass fit_func: myname
    in the mead config if you want your own bypass, in which case training is entirely delegate to the 3rd party code.

    This use-case is expected to be extremely uncommon.  More common behavior would be to override the Trainer and use
    the provided fit function.

    :param model:
    :param ts:
    :param vs:
    :param es:
    :param kwargs:
    :return:
    """
    if type(model_params) is dict:
        task_name = model_params['task']
    else:
        task_name = model_params.task_name
    fit_func_name = kwargs.get('fit_func', 'default')
    return BASELINE_FIT_FUNC[task_name][fit_func_name](model_params, ts, vs, es, **kwargs)


@export
def create_trainer(model_params, **kwargs):
    """Create the default trainer, or a user-defined one if `trainer_type` is not `default`

    :param default_create_model_fn: The constructor for the default trainer (defined in each platform/task)
    :param model: The model to train
    :param kwargs:
    :return:
    """
    trainer_type = kwargs.get('trainer_type', 'default')
    if type(model_params) is dict:
        task_name = model_params['task']
    else:
        task_name = model_params.task_name
    Constructor = BASELINE_TRAINERS[task_name][trainer_type]
    return Constructor(model_params, **kwargs)


def calc_lr_params(train_params, num_steps):
    # This belongs in the trainer!
    if train_params.get('lr_scheduler_type', None) == 'zaremba':
        first_range = int(train_params['start_decay_epoch'] * num_steps)
        train_params['boundaries'] = [first_range] + list(
            np.arange(
                train_params['start_decay_epoch'] + 1,
                train_params['epochs'] + 1,
                dtype=np.int32
            ) * num_steps
        )



import os
import re
import ast
import shutil
from itertools import product
from setuptools import setup, find_packages


def get_version(file_name, version_name="__version__"):
    with open(file_name) as f:
        tree = ast.parse(f.read())
        for node in ast.walk(tree):
            if isinstance(node, ast.Assign):
                if node.targets[0].id == version_name:
                    return node.value.s
    raise ValueError(f"Unable to find an assignment to the variable {version_name} in file {file_name}")


class About(object):
    NAME = 'baseline'
    AUTHOR = 'dpressel'
    VERSION = get_version('baseline/version.py')
    EMAIL = "{}@gmail.com".format(AUTHOR)
    URL = "https://www.github.com/{}/{}".format(AUTHOR, NAME)
    DOWNLOAD_URL = "{}/archive/{}.tar.gz".format(URL, VERSION)
    DOC_URL = "{}/tree/master/".format(URL)
    DOC_NAME = 'docs/{}.md'.format(NAME)

def get_configs(config_loc):
    """include everything in config_loc as package data."""
    configs = []
    for f in os.listdir(config_loc):
        configs.append(os.path.join(config_loc, f))
    write_manifest(configs)
    return configs

def write_manifest(lines):
    with open("MANIFEST.in", "w") as f:
        f.write("\n".join(map(lambda x: 'include {}'.format(x), lines)))

def fix_links(text):
    """Pypi doesn't seem to host multiple docs so replace local links with ones to github."""
    regex = re.compile(r"\[(.*?)\]\(((?:docs|docker)/.*?\.md)\)")
    text = regex.sub(r"[\1]({}\2)".format(About.DOC_URL), text)
    return text

def read_doc(f_name, new_name=None, fix_fn=fix_links):
    """
    Because our readme is outside of this dir we need to copy it in so
    that it is picked up by the install.
    """
    if new_name is None:
        new_name = f_name
    path = os.path.dirname(os.path.realpath(__file__))
    doc_loc = os.path.normpath(os.path.join(path, '..', f_name))
    new_loc = os.path.join(path, new_name)
    if os.path.isfile(doc_loc):
        shutil.copyfile(doc_loc, new_loc)
    descript = open(new_loc, 'r').read()
    return fix_fn(descript)

def main():
    setup(
        name='mead-{}'.format(About.NAME),
        version=About.VERSION,
        description='Strong Deep-Learning Baseline algorithms for NLP',
        long_description=read_doc(About.DOC_NAME, new_name='README.md'),
        long_description_content_type="text/markdown",
        author=About.AUTHOR,
        author_email=About.EMAIL,
        license='Apache 2.0',
        url=About.URL,
        download_url=About.DOWNLOAD_URL,
        packages=find_packages(exclude=['tests', 'layers*', 'api-examples']),
        package_data={
            'mead': get_configs('mead/config'),
        },
        include_package_data=True,
        install_requires=[
            'numpy',
            'six',
            'mead-layers=={}'.format(About.VERSION),
        ],
        extras_require={
            'test': ['pytest', 'mock', 'contextdecorator', 'pytest-forked'],
            'report': ['visdom', 'tensorboardX'],
            'yaml': ['pyyaml'],
            'tf2': ['tensorflow_addons'],
            'grpc': ['grpc'],
            'onnx': ['onnxruntime']
        },
        entry_points={
            'console_scripts': [
                'mead-train = mead.trainer:main',
                'mead-export = mead.export:main',
                'mead-clean = mead.clean:main',
                'mead-eval = mead.eval:main',
                'bleu = baseline.bleu:main',
                'conlleval = baseline.conlleval:main',
            ]
        },
        classifiers={
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        },
        keywords=['deep-learning', 'nlp', 'pytorch', 'tensorflow'],
    )

if __name__ == "__main__":
    main()



from baseline.reader import CONLLSeqReader
import numpy as np
from collections import Counter
import codecs
import io
import os
import re
import baseline.data


class CONLLSeqMixedCaseReader(CONLLSeqReader):

    def __init__(self, max_sentence_length=-1, max_word_length=-1, word_trans_fn=None,
                 vec_alloc=np.zeros, vec_shape=np.shape, trim=False, extended_features=dict()):
        super(CONLLSeqMixedCaseReader, self).__init__(max_sentence_length, max_word_length, word_trans_fn, vec_alloc, vec_shape, trim, extended_features)
        self.idx = 2 # GO=0, START=1, EOS=2


    def load(self, filename, vocabs, batchsz, shuffle=False, do_sort=True):
        ts = []
        words_vocab = vocabs['word']
        chars_vocab = vocabs['char']
        mxlen = self.max_sentence_length
        maxw = self.max_word_length
        extracted = self.read_examples(filename)
        texts = extracted['texts']
        labels = extracted['labels']

        for i in range(len(texts)):

            xs_ch = self.vec_alloc((mxlen, maxw), dtype=np.int)
            xs = self.vec_alloc((mxlen), dtype=np.int)
            xs_lc = self.vec_alloc((mxlen), dtype=np.int)
            ys = self.vec_alloc((mxlen), dtype=np.int)
            keys = self.extended_features.keys()

            item = {}
            for key in keys:
                item[key] = self.vec_alloc((mxlen), dtype=np.int)

            text = texts[i]
            lv = labels[i]

            length = mxlen
            for j in range(mxlen):

                if j == len(text):
                    length = j
                    break

                w = text[j]
                nch = min(len(w), maxw)
                label = lv[j]

                if label not in self.label2index:
                    self.idx += 1
                    self.label2index[label] = self.idx

                ys[j] = self.label2index[label]
                xs_lc[j] = words_vocab.get(w.lower(), 0)
                xs[j] = words_vocab.get(w, 0)
                # Extended features
                for key in keys:
                    item[key][j] = vocabs[key].get(extracted[key][i][j])
                for k in range(nch):
                    xs_ch[j, k] = chars_vocab.get(w[k], 0)
            item.update({'x': xs, 'x_lc': xs_lc, 'xch': xs_ch, 'y': ys, 'lengths': length, 'ids': i})
            ts.append(item)
        examples = baseline.data.SeqWordCharTagExamples(ts, do_shuffle=shuffle, do_sort=do_sort)
        return baseline.data.SeqWordCharLabelDataFeed(examples, batchsz=batchsz, shuffle=shuffle,
                                                      vec_alloc=self.vec_alloc, vec_shape=self.vec_shape), texts

    def build_vocab(self, files):
        vocab_word = Counter()
        vocab_ch = Counter()
        vocab_word['<UNK>'] = 1
        vocabs = {}
        keys = self.extended_features.keys()
        for key in keys:
            vocabs[key] = Counter()

        maxw = 0
        maxs = 0
        for file in files:
            if file is None:
                continue

            sl = 0
            with codecs.open(file, encoding='utf-8', mode='r') as f:
                for line in f:

                    line = line.strip()
                    if line == '':
                        maxs = max(maxs, sl)
                        sl = 0

                    else:
                        states = re.split("\s", line)
                        sl += 1
                        w = states[0]
                        vocab_word[w] += 1
                        vocab_word[w.lower()] += 1
                        maxw = max(maxw, len(w))
                        for k in w:
                            vocab_ch[k] += 1
                        for key, index in self.extended_features.items():
                            vocabs[key][states[index]] += 1

        self.max_word_length = min(maxw, self.max_word_length) if self.max_word_length > 0 else maxw
        self.max_sentence_length = min(maxs, self.max_sentence_length) if self.max_sentence_length > 0 else maxs
        print('Max sentence length %d' % self.max_sentence_length)
        print('Max word length %d' % self.max_word_length)

        vocabs.update({'char': vocab_ch, 'word': vocab_word})
        return vocabs


def create_seq_pred_reader(mxlen, mxwlen, word_trans_fn, vec_alloc, vec_shape, trim, **kwargs):
    return CONLLSeqMixedCaseReader(mxlen, mxwlen, word_trans_fn, vec_alloc, vec_shape, trim)



from baseline.pytorch.torchy import *
from baseline.pytorch.transformer import TransformerEncoderStack, subsequent_mask, MultiHeadedAttention
from baseline.model import LanguageModel, register_model
import torch.autograd
import os


class LanguageModelBase(nn.Module, LanguageModel):
    def __init__(self):
        super().__init__()

    def save(self, outname):
        torch.save(self, outname)
        basename, _ = os.path.splitext(outname)

    def create_loss(self):
        return SequenceCriterion(LossFn=nn.CrossEntropyLoss)

    @classmethod
    def load(cls, filename, **kwargs):
        device = kwargs.get('device')
        if not os.path.exists(filename):
            filename += '.pyt'
        model = torch.load(filename, map_location=device)
        model.gpu = False if device == 'cpu' else model.gpu
        return model

    def zero_state(self, batchsz):
        return None

    @property
    def requires_state(self):
        pass

    def make_input(self, batch_dict, numpy_to_tensor=False):
        example_dict = dict({})
        for key in self.src_keys:

            tensor = batch_dict[key]
            if numpy_to_tensor:
                tensor = torch.from_numpy(tensor)

            if self.gpu:
                tensor = tensor.cuda()

            example_dict[key] = tensor

        y = batch_dict.get('y')
        if y is not None:
            if numpy_to_tensor:
                y = torch.from_numpy(y)
            if self.gpu:
                y = y.cuda()
            example_dict['y'] = y
        return example_dict

    @classmethod
    def create(cls, embeddings, **kwargs):

        lm = cls()
        lm.gpu = kwargs.get('gpu', True)
        lm.tgt_key = kwargs.get('tgt_key')
        if lm.tgt_key is None:
            raise Exception('Need a `tgt_key` to know which source vocabulary should be used for destination ')

        lm.src_keys = kwargs.get('src_keys', embeddings.keys())
        lm.create_layers(embeddings, **kwargs)
        return lm

    def create_layers(self, embeddings, **kwargs):
        """This method defines the model itself, and must be overloaded by derived classes

        This function will update `self` with the layers required to execute the `call()` method

        :param embeddings: The input feature indices
        :param kwargs:
        :return:
        """

    def predict(self, batch_dict, **kwargs):
        numpy_to_tensor = bool(kwargs.get('numpy_to_tensor', True))
        batch_dict = self.make_input(batch_dict, numpy_to_tensor=numpy_to_tensor)
        hidden = batch_dict.get('h')
        step_softmax, _ = self(batch_dict, hidden)
        return F.softmax(step_softmax, dim=-1)


class AbstractGeneratorLanguageModel(LanguageModelBase):

    def create_layers(self, embeddings, **kwargs):
        self.embeddings = self.init_embed(embeddings, **kwargs)
        self.embeddings_proj = self.init_embeddings_proj(**kwargs)
        self.generator = self.init_generate(**kwargs)
        self.output_layer = self.init_output(embeddings, **kwargs)

    def forward(self, input: Dict[str, TensorDef], hidden: TensorDef) -> Tuple[TensorDef, TensorDef]:
        emb = self.embed(input)
        output, hidden = self.generate(emb, hidden)
        return self.output_layer(output), hidden

    def embed(self, input):
        embedded_dropout = self.embeddings(input)
        return self.embeddings_proj(embedded_dropout)

    def init_embed(self, embeddings: Dict[str, TensorDef], **kwargs) -> BaseLayer:
        """This method creates the "embedding" layer of the inputs, with an optional reduction

        :param embeddings: A dictionary of embeddings

        :Keyword Arguments: See below
        * *embeddings_reduction* (defaults to `concat`) An operator to perform on a stack of embeddings
        * *embeddings_dropout = float(kwargs.get('embeddings_dropout', 0.0))

        :return: The output of the embedding stack followed by its reduction.  This will typically be an output
          with an additional dimension which is the hidden representation of the input
        """
        reduction = kwargs.get('embeddings_reduction', 'concat')
        embeddings_dropout = float(kwargs.get('embeddings_dropout', 0.0))
        return EmbeddingsStack({k: embeddings[k] for k in self.src_keys}, embeddings_dropout, reduction=reduction)

    def init_embeddings_proj(self, **kwargs):
        input_sz = self.embeddings.output_dim
        hsz = kwargs.get('hsz', kwargs.get('d_model'))
        if hsz != input_sz:
            proj = pytorch_linear(input_sz, hsz)
            print('Applying a transform from {} to {}'.format(input_sz, hsz))
        else:
            proj = nn.Identity()
        return proj

    def init_generate(self, **kwargs):
        pass

    def generate(self, emb, hidden):
        return self.generator((emb, hidden))

    def init_output(self, embeddings, **kwargs):
        self.vsz = embeddings[self.tgt_key].get_vsz()
        hsz = kwargs.get('hsz', kwargs.get('d_model'))
        unif = float(kwargs.get('unif', 0.0))
        do_weight_tying = bool(kwargs.get('tie_weights', False))
        output_bias = kwargs.get('output_bias', False)
        if do_weight_tying:
            output = WeightTieDense(embeddings[self.tgt_key], output_bias)
        else:
            output = pytorch_linear(hsz, self.vsz, unif)
        return output


@register_model(task='lm', name='default')
class RNNLanguageModel(AbstractGeneratorLanguageModel):

    def __init__(self):
        super().__init__()

    def zero_state(self, batchsz):
        weight = next(self.parameters()).data
        return (torch.autograd.Variable(weight.new(self.num_layers, batchsz, self.hsz).zero_()),
                torch.autograd.Variable(weight.new(self.num_layers, batchsz, self.hsz).zero_()))

    @property
    def requires_state(self):
        True

    def init_generate(self, **kwargs):
        pdrop = float(kwargs.get('dropout', 0.5))
        self.num_layers = kwargs.get('layers', kwargs.get('num_layers', 1))
        self.hsz = kwargs.get('hsz', kwargs.get('d_model'))
        return WithDropoutOnFirst(LSTMEncoderWithState(self.hsz, self.hsz, self.num_layers, pdrop, batch_first=True),
                                  pdrop,
                                  kwargs.get('variational', False))


@register_model(task='lm', name='transformer')
class TransformerLanguageModel(AbstractGeneratorLanguageModel):

    def __init__(self):
        super().__init__()

    @property
    def requires_state(self):
        False

    def init_layer_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding, nn.LayerNorm)):
            module.weight.data.normal_(mean=0.0, std=self.weight_std)
        if isinstance(module, (nn.Linear, nn.LayerNorm)) and module.bias is not None:
            module.bias.data.zero_()

    def init_generate(self, **kwargs):
        pdrop = float(kwargs.get('dropout', 0.1))
        layers = kwargs.get('layers', kwargs.get('num_layers', 1))
        d_model = int(kwargs.get('d_model', kwargs.get('hsz')))
        num_heads = kwargs.get('num_heads', 4)
        d_ff = int(kwargs.get('d_ff', 4 * d_model))
        rpr_k = kwargs.get('rpr_k')
        d_k = kwargs.get('d_k')
        scale = bool(kwargs.get('scale', True))
        activation = kwargs.get('activation', 'gelu')
        layer_norm_eps = kwargs.get('layer_norm_eps', 1e-12)
        layer_norms_after = kwargs.get('layer_norms_after', False)
        return TransformerEncoderStack(num_heads, d_model=d_model, pdrop=pdrop, scale=scale,
                                       layers=layers, d_ff=d_ff, rpr_k=rpr_k, d_k=d_k,
                                       activation=activation, layer_norm_eps=layer_norm_eps, layer_norms_after=layer_norms_after)

    def create_layers(self, embeddings, **kwargs):
        super().create_layers(embeddings, **kwargs)
        self.weight_std = kwargs.get('weight_std', 0.02)
        self.apply(self.init_layer_weights)

    def create_mask(self, bth):
        T = bth.shape[1]
        mask = subsequent_mask(T).type_as(bth)
        return mask

    def generate(self, bth, _):
        mask = self.create_mask(bth)
        return self.generator((bth, mask)), None


@register_model(task='lm', name='transformer-mlm')
class TransformerMaskedLanguageModel(TransformerLanguageModel):

    def create_mask(self, bth):
        T = bth.shape[1]
        mask = torch.ones((1, 1, T, T)).type_as(bth)
        return mask



import time
import logging
from baseline.pytorch.torchy import *
from eight_mile.utils import listify, revlut
from eight_mile.pytorch.optz import OptimizerManager
from baseline.utils import get_model_file, get_metric_cmp
from baseline.train import Trainer, create_trainer, register_trainer, register_training_func
from baseline.model import create_model_for
from torch.utils.data import DataLoader
logger = logging.getLogger('baseline')


@register_trainer(task='lm', name='default')
class LanguageModelTrainerPyTorch(Trainer):

    def __init__(self, model, **kwargs):
        super().__init__()
        if type(model) is dict:
            model = create_model_for('lm', **model)
        self.model = model
        self.clip = float(kwargs.get('clip', 5))
        self.gpus = kwargs.get('gpus', 1)
        if self.gpus > 0:
            self.crit = model.create_loss().cuda()
            if self.gpus > 1:
                self.model = torch.nn.DataParallel(model).cuda()
            else:
                self.model.cuda()
        else:
            logger.warning("Requested training on CPU.  This will be slow.")
            self.crit = model.create_loss()

        self.nsteps = kwargs.get('nsteps', 500)
        self.optimizer = OptimizerManager(self.model, **kwargs)

    def repackage_hidden(self, h):
        """Wraps hidden states in new Variables, to detach them from their history."""
        if isinstance(h, torch.Tensor):
            return h.detach()
        else:
            return tuple(self.repackage_hidden(v) for v in h)

    def save(self, model_file):
        self._get_pytorch_model().save(model_file)

    def _get_pytorch_model(self):
        return self.model.module if self.gpus > 1 else self.model

    @staticmethod
    def _get_dims(loader):
        batch_dict = loader.dataset[0]
        return batch_dict['y'].shape

    @staticmethod
    def _num_toks(batch_dict):
        return np.prod(batch_dict['y'].shape)

    def calc_metrics(self, agg, norm):
        metrics = super().calc_metrics(agg, norm)
        metrics['perplexity'] = np.exp(metrics['avg_loss'])
        return metrics

    def test(self, vs, reporting_fns, phase='Valid', **kwargs):
        epoch = 0
        if phase == 'Valid':
            self.valid_epochs += 1
            epoch = self.valid_epochs
        start = time.time()
        self.model.eval()
        total_loss = 0
        total_toks = 0
        batchsz, nctx = self._get_dims(vs)
        hidden = self._get_pytorch_model().zero_state(batchsz)

        for batch_dict in vs:
            inputs = self._get_pytorch_model().make_input(batch_dict)
            y = inputs.pop('y')
            output, hidden = self.model(inputs, hidden)
            toks = self._num_toks(batch_dict)
            total_loss += self.crit(output, y).item() * toks
            total_toks += toks
            if hidden is not None:
                hidden = self.repackage_hidden(hidden)
        metrics = self.calc_metrics(total_loss, total_toks)
        self.report(
            epoch, metrics, start,
            phase, 'EPOCH', reporting_fns
        )
        return metrics

    def train(self, ts, reporting_fns):
        start = time.time()
        self.nstep_start = start
        self.model.train()
        epoch_loss = 0
        epoch_toks = 0
        batchsz, nctx = self._get_dims(ts)
        hidden = self._get_pytorch_model().zero_state(batchsz)

        for batch_dict in ts:
            if hidden is not None:
                hidden = self.repackage_hidden(hidden)
            inputs = self._get_pytorch_model().make_input(batch_dict)
            y = inputs.pop('y')
            self.optimizer.zero_grad()
            output, hidden = self.model(inputs, hidden)
            loss = self.crit(output, y)
            loss.backward()
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.clip)
            self.optimizer.step()
            toks = self._num_toks(batch_dict)
            report_loss = loss.item() * toks
            epoch_loss += report_loss
            epoch_toks += toks
            self.nstep_agg += report_loss
            self.nstep_div += toks
            if (self.optimizer.global_step + 1) % self.nsteps == 0:
                metrics = self.calc_metrics(self.nstep_agg, self.nstep_div)
                self.report(
                    self.optimizer.global_step + 1, metrics, self.nstep_start,
                    'Train', 'STEP', reporting_fns, self.nsteps
                )
                self.reset_nstep()

        metrics = self.calc_metrics(epoch_loss, epoch_toks)
        self.train_epochs += 1
        self.report(
            self.train_epochs, metrics, start,
            'Train', 'EPOCH', reporting_fns
        )
        return metrics


@register_training_func('lm')
def fit(model, ts, vs, es, **kwargs):
    epochs = int(kwargs['epochs']) if 'epochs' in kwargs else 5
    patience = int(kwargs['patience']) if 'patience' in kwargs else epochs
    do_early_stopping = bool(kwargs.get('do_early_stopping', True))
    model_file = get_model_file('lm', 'pytorch', kwargs.get('basedir'))

    num_loader_workers = int(kwargs.get('num_loader_workers', 0))
    pin_memory = bool(kwargs.get('pin_memory', True))
    ts = DataLoader(ts, num_workers=num_loader_workers, batch_size=None, pin_memory=pin_memory)
    vs = DataLoader(vs, batch_size=None, pin_memory=pin_memory)
    es = DataLoader(es, batch_size=None, pin_memory=pin_memory) if es is not None else None

    best_metric = 10000
    if do_early_stopping:
        early_stopping_metric = kwargs.get('early_stopping_metric', 'avg_loss')
        early_stopping_cmp, best_metric = get_metric_cmp(early_stopping_metric, kwargs.get('early_stopping_cmp'))
        patience = kwargs.get('patience', epochs)
        logger.info('Doing early stopping on [%s] with patience [%d]', early_stopping_metric, patience)

    reporting_fns = listify(kwargs.get('reporting', []))
    logger.info('reporting %s', reporting_fns)

    after_train_fn = kwargs.get('after_train_fn', None)
    trainer = create_trainer(model, **kwargs)
    last_improved = 0
    for epoch in range(epochs):

        trainer.train(ts, reporting_fns)
        if after_train_fn is not None:
            after_train_fn(model)
        test_metrics = trainer.test(vs, reporting_fns, phase='Valid')

        if do_early_stopping is False:
            trainer.save(model_file)

        elif early_stopping_cmp(test_metrics[early_stopping_metric], best_metric):
            last_improved = epoch
            best_metric = test_metrics[early_stopping_metric]
            logger.info('New best %.3f', best_metric)
            trainer.save(model_file)

        elif (epoch - last_improved) > patience:
            logger.info('Stopping due to persistent failures to improve')
            break

    if do_early_stopping is True:
        logger.info('Best performance on %s: %.3f at epoch %d', early_stopping_metric, best_metric, last_improved)

    if es is not None:
        logger.info('Reloading best checkpoint')
        model = torch.load(model_file)
        trainer = create_trainer(model, **kwargs)
        test_metrics = trainer.test(es, reporting_fns, phase='Test')
    return test_metrics



import os
import json
import shutil
import logging
import logging.config
import hashlib
from typing import Optional, List, Dict
from datetime import datetime
import argparse
from copy import deepcopy
from itertools import chain
from collections import OrderedDict, MutableMapping
from baseline.utils import exporter, str2bool, read_config_file, write_json, get_logging_level, validate_url, zip_files, delete_old_copy

__all__ = []
export = exporter(__all__)
logger = logging.getLogger('mead')

KNOWN_DATE_FMTS = [
    '%Y%m%d',
    '%Y-%m-%d',
    '%Y/%m/%d',
    '%Y',
    '%Y%m',
    '%Y-%m',
    '%Y/%m',
    '%Y%m%d_%H%M',
    '%Y-%m-%d_%H-%M',
    '%Y/%m/%d_%M/%H'
]


@export
def configure_logger(logger_config, basedir=None):
    """Use the logger file (logging.json) to configure the log, but overwrite the filename to include the PID

    There are reporting and timing loggers that are configured, the latter being used for speed testing.

    :param logger_config: The logging configuration JSON or file containing JSON
    :return: A dictionary config derived from the logger_file, with the reporting handler suffixed with PID
    """

    config = read_config_file_or_json(logger_config, 'logger')
    config['handlers']['reporting_file_handler']['filename'] = 'reporting-{}.log'.format(os.getpid())
    config['handlers']['timing_file_handler']['filename'] = 'timing-{}.log'.format(os.getpid())
    if basedir is not None:
        # Create a copy of the file handler so we can keep the log in cwd for back compat
        config['handlers']['reporting_basedir_handler'] = deepcopy(config['handlers']['reporting_file_handler'])
        # Update the file to be in the basedir of the model
        config['handlers']['reporting_basedir_handler']['filename'] = os.path.join(basedir, config['handlers']['reporting_basedir_handler']['filename'])
        config['handlers']['reporting_basedir_handler']['class'] = 'baseline.utils.MakeFileHandler'
        # Add this handler to the logger
        config['loggers']['baseline.reporting']['handlers'].append('reporting_basedir_handler')
        # Same as above
        config['handlers']['timing_basedir_handler'] = deepcopy(config['handlers']['timing_file_handler'])
        config['handlers']['timing_basedir_handler']['filename'] = os.path.join(basedir, config['handlers']['timing_basedir_handler']['filename'])
        config['handlers']['timing_basedir_handler']['class'] = 'baseline.utils.MakeFileHandler'
        config['loggers']['baseline.timing']['handlers'].append('timing_basedir_handler')
    level = os.getenv('LOG_LEVEL', 'INFO')
    config['loggers']['baseline']['level'] = get_logging_level(os.getenv('BASELINE_LOG_LEVEL', level))
    config['loggers']['mead']['level'] = get_logging_level(os.getenv('MEAD_LOG_LEVEL', level))
    config['handlers']['reporting_console_handler']['level'] = get_logging_level(os.getenv('REPORTING_LOG_LEVEL', level))
    config['handlers']['timing_console_handler']['level'] = get_logging_level(os.getenv('TIMING_LOG_LEVEL', level))

    logging.config.dictConfig(config)


@export
def print_dataset_info(dataset):
    logger.info("[train file]: {}".format(dataset['train_file']))
    logger.info("[valid file]: {}".format(dataset['valid_file']))
    if 'test_file' in dataset:
        logger.info("[test file]: {}".format(dataset['test_file']))
    vocab_file = dataset.get('vocab_file')
    if vocab_file is not None:
        logger.info("[vocab file]: {}".format(vocab_file))
    label_file = dataset.get('label_file')
    if label_file is not None:
        logger.info("[label file]: {}".format(label_file))


@export
def read_config_file_or_json(config, name=''):
    if isinstance(config, (dict, list)):
        return config
    config = os.path.expanduser(config)
    if os.path.exists(config):
        return read_config_file(config)
    raise Exception('Expected {} config file or a JSON object.'.format(name))


def parse_date(s, known_fmts: List[str] = KNOWN_DATE_FMTS):
    for fmt in known_fmts:
        try:
            dt = datetime.strptime(s, fmt)
            return dt
        except:
            continue
    raise Exception("Couldn't parse datestamp {}".format(s))


@export
def flatten(dictionary, sep='.'):
    """Flatten a nested dict.

    :param dictionary: the dictionary to flatten.
    :param sep: The separator between old nested keys.
    :returns: The flattened dict.
    """

    def _flatten(dictionary, sep, prev):
        """This is the recursive function that actually does the work of flattening..

        :param dictionary: the dictionary to flatten.
        :param sep: The separator between old nested keys.
        :param prev: A recurrent state param that tracks key above you.
        :returns: The flattened dict.
        """
        flat = {}
        prev = [] if prev is None else prev
        for k, v in dictionary.items():
            if isinstance(v, MutableMapping):
                flat.update(_flatten(v, sep=sep, prev=list(chain(prev, [k]))))
            else:
                flat[sep.join(chain(prev, [k]))] = v
        return flat

    return _flatten(dictionary, sep, [])


@export
def unflatten(dictionary, sep: str = "."):
    """Turn a flattened dict into a nested dict.

    :param dictionary: The dict to unflatten.
    :param sep: This character represents a nesting level in a flattened key
    :returns: The nested dict
    """
    nested = {}
    for k, v in dictionary.items():
        keys = k.split(sep)
        it = nested
        for key in keys[:-1]:
            # If key is in `it` we get the value otherwise we get a new dict.
            # .setdefault will also set the new dict to the value of `it[key]`.
            # assigning to `it` will move us a step deeper into the nested dict.
            it = it.setdefault(key, {})
        it[keys[-1]] = v
    return nested


def _infer_numeric_or_str(value: str):
    """Convert value to an int, float or leave it as a string.

    :param value: The cli value to parse
    :returns: The value converted to int or float if possible
    """
    for func in (int, float):
        try:
            return func(value)
        except ValueError:
            continue
    return value


@export
def parse_overrides(overrides, pre):
    """Find override parameters in the cli args.

    Note:
        If you use the same cli flag multiple time the values will be
        aggregated into a list. For example
        `--x:a 1 --x:a 2 --x:a 3` will give back `{'a': [1, 2, 3]}`

    :param overrides: The cli flags and values.
    :param pre: only look at keys that start with `--{pre}:`
    :returns: The key value pairs from the command line args.
    """
    pre = f'--{pre}:'
    parser = argparse.ArgumentParser()
    for key in set(filter(lambda x: pre in x, overrides)):
        # Append action collect each value into a list allowing us to override a
        # yaml list by repeating the key.
        parser.add_argument(key, action='append', type=_infer_numeric_or_str)
    args = parser.parse_known_args(overrides)[0]
    return {k.split(":")[1]: v[0] if len(v) == 1 else v for k, v in vars(args).items()}


@export
def parse_and_merge_overrides(base, overrides, pre):
    """Parse extra cli args and use them to override the original config.

    :param base: The base config that will be overridden
    :param overrides: The cli args holding override values
    :param pre: The key used to find cli flags.
    :returns: The base config overridden with values from overrides.
    """
    overrides = parse_overrides(overrides, pre)
    base = flatten(base)
    base.update(overrides)
    return unflatten(base)


@export
def get_dataset_from_key(dataset_key, datasets_set):

    # This is the previous behavior
    if dataset_key in datasets_set:
        return datasets_set[dataset_key]

    last_date = parse_date('1900')
    last_k = None
    for k, v in datasets_set.items():

        if dataset_key in k:
            try:
                parts = k.split(':')
                dt = parse_date(parts[-1])
                dataset_name = ":".join(parts[:-1])
            except Exception:
                continue
            if dt > last_date and dataset_name == dataset_key:
                last_date = dt
                last_k = k

    if last_k is None:
        raise Exception("No dataset could be found with key {}".format(dataset_key))

    return datasets_set[last_k]


@export
def get_mead_settings(mead_settings_config):
    if mead_settings_config is None:
        return {}
    return read_config_file_or_json(mead_settings_config, 'mead settings')


@export
def index_by_label(object_list):
    objects = {x['label']: x for x in object_list}
    return objects


@export
def convert_path(path, loc=None):
    """If the provided path doesn't exist search for it relative to loc (or this file)."""
    if os.path.isfile(path) or path.startswith("$") or validate_url(path):
        return path
    if path.startswith("http"):
        return path
    if path.startswith("hub:"):
        return path
    if loc is None:
        loc = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(loc, path)


def _infer_type_or_str(x):
    try:
        return str2bool(x)
    except:
        try:
            return int(x)
        except ValueError:
            try:
                return float(x)
            except ValueError:
                return x


@export
def parse_extra_args(base_args, extra_args):
    """Parse extra command line arguments based on based names.
    Note:
        special args should be in the form --{base_name}:{special_name}
    :param base_args: List[str], A list of base argument names.
    :param extra_args: List[str], A list of special arguments and values.
    :returns:
        dict, The parsed special settings in the form
        {
            "base": {
                "special": val,
                ...
            },
            ...
        }
    """
    found_args = []
    for arg in base_args:
        key = "{}:".format(arg)
        for extra_arg in extra_args:
            if key in extra_arg:
                found_args.append(extra_arg)
    parser = argparse.ArgumentParser()
    for key in found_args:
        parser.add_argument(key, type=_infer_type_or_str)
    args = parser.parse_known_args(extra_args)[0]
    settings = {arg: {} for arg in base_args}
    args = vars(args)
    for key in args:
        base, extra = key.split(":")
        settings[base][extra] = args[key]
    return settings


@export
def order_json(data):
    """Sort json to a consistent order.
    When you hash json that has the some content but is different orders you get
    different fingerprints.
    In:  hashlib.sha1(json.dumps({'a': 12, 'b':14}).encode('utf-8')).hexdigest()
    Out: '647aa7508f72ece3f8b9df986a206d95fd9a2caf'
    In:  hashlib.sha1(json.dumps({'b': 14, 'a':12}).encode('utf-8')).hexdigest()
    Out: 'a22215982dc0e53617be08de7ba9f1a80d232b23'
    This function sorts json by key so that hashes are consistent.
    Note:
        In our configs we only have lists where the order doesn't matter so we
        can sort them for consistency. This would have to change if we add a
        config field that needs order we will need to refactor this.
    :param data: dict, The json data.
    :returns:
        collections.OrderedDict: The data in a consistent order (keys sorted alphabetically).
    """
    new = OrderedDict()
    for (key, value) in sorted(data.items(), key=lambda x: x[0]):
        if isinstance(value, dict):
            value = order_json(value)
        new[key] = value
    return new


KEYS = {
    ('conll_output',),
    ('visdom',),
    ('visdom_name',),
    ('model', 'gpus'),
    ('test_thresh',),
    ('reporting',),
    ('num_valid_to_show',),
    ('train', 'verbose'),
    ('train', 'model_base'),
    ('train', 'model_zip'),
    ('train', 'nsteps'),
    ('test_batchsz',),
    ('basedir',),
}


@export
def remove_extra_keys(config, keys=KEYS):
    """Remove config items that don't effect the model.
    We base most things off of the sha1 hash of the model configs but there
    is a problem. Some things in the config file don't effect the model such
    as the name of the `conll_output` file or if you are using `visdom`
    reporting. This strips out these kind of things so that as long as the model
    parameters match the sha1 will too.
    :param config: dict, The json data.
    :param keys: Set[Tuple[str]], The keys to remove.
    :returns:
        dict, The config with certain keys removed.
    """
    c = deepcopy(config)
    for key in keys:
        x = c
        for k in key[:-1]:
            x = x.get(k)
            if x is None:
                break
        else:
            _ = x.pop(key[-1], None)
    return c


@export
def hash_config(config):
    """Hash a json config with sha1.
    :param config: dict, The config to hash.
    :returns:
        str, The sha1 hash.
    """
    stripped_config = remove_extra_keys(config)
    sorted_config = order_json(stripped_config)
    json_bytes = json.dumps(sorted_config).encode('utf-8')
    return hashlib.sha1(json_bytes).hexdigest()


def _listdir(model_dir):
    try:
        return os.listdir(model_dir)
    except OSError:
        return []


@export
def find_model_version(model_dir):
    """Find the next usable model version when exporting.

    :param model_dir: `str` The directory we are exporting to.

    :returns: `str` The model version.
    """
    return str(max(chain([0], map(int, filter(lambda x: x.isdigit(), _listdir(model_dir))))) + 1)


@export
def get_output_paths(
        output_dir: str,
        project: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        remote: bool = False,
        make_server: bool = True,
        use_version: bool = True,
    ):
    """Create the output paths for export.

    Old behavior:
    if remote == True:
        f"{output_dir}/client/{basename(output_dir)}/{version}"
        f"{output_dir}/server/{basename(output_dir)}/{version}"
    else:
        f"{output_dir}/{version}"

    New Behavior:
    if remote == True:
        f"{output_dir}/client/{project}/{name}/{version}"
        f"{output_dir}/server/{project}/{name}/{version}"
    else:
        f"{output_dir}/{project}/{name}/{version}"

    If either project or name is None then they are skipped in the path.

    If you want to create output that looks the same as the old versions
    then set both project and name to None. This will skip them if remote is
    false or it will add the basename of output_dir to the path if remote.

    :param output_dir: `str` The base of these paths.
    :param project: `str` The first subdir in the created path.
    :param name: `str` The second subdir in the created path.
    :param version: `str` The model version.
    :param remote: `bool` Should we create separate client and server bundles?
    :param maker_server: `bool` When false don't actually make the directory that the
        server part will go in. This is because TF really wants to make it itself.
    :param use_version: Should we use
    :returns: `Tuple[str, str]` The client and server output dirs.
    """
    # In this case we use the basename to simulate the old behavior.
    if remote and project is None and name is None:
        project = os.path.basename(output_dir)
    project = project if project is not None else ''
    name = name if name is not None else ''
    client = 'client' if remote else ''
    server = 'server' if remote else ''
    server_path = [output_dir, server, project, name]
    client_path = [output_dir, client, project, name]
    # Sniff the dir and see what version we should use
    if use_version:
        version = find_model_version(os.path.join(*server_path)) if version is None else version
        server_path.append(version)
        client_path.append(version)
    server_path = os.path.join(*server_path)
    client_path = os.path.join(*client_path)
    if remote:
        os.makedirs(client_path)
    if make_server:
        os.makedirs(server_path)
    return client_path, server_path


@export
def get_export_params(
        config: Dict,
        output_dir: Optional[str] = None,
        project: Optional[str] = None,
        name: Optional[str] = None,
        model_version: Optional[str] = None,
        exporter_type: Optional[str] = None,
        return_labels: Optional[bool] = None,
        is_remote: Optional[bool] = None
):
    """Combine export parameters from the config file and cli arguments.

    :param config: The export block of the config.
    :param output_dir: The base of export paths. (defaults to './models')
    :param project: The name of the project this model is for.
    :param name: The name of this model (often the use case for it, `ner`, `intent` etc).
    :param model_version: The version of this model.
    :param exporter_type: The name of the exporter to use (defaults to 'default')
    :param return_labels: Should labels be returned? (defaults to False)
    :param is_remote: Should the bundle be split into client and server dirs.
    :param user_version

    :returns: `Tuple[str, str, str, str, str, bool, bool]`
        The output_dir, project, name, model_version, exporter_type, return_labels, and remote
    """
    project = project if project is not None else config.get('project')
    name = name if name is not None else config.get('name')
    output_dir = output_dir if output_dir is not None else config.get('output_dir', './models')
    output_dir = os.path.expanduser(output_dir)
    model_version = model_version if model_version is not None else config.get('model_version')
    exporter_type = exporter_type if exporter_type is not None else config.get('type', config.get('exporter_type', 'default'))
    return_labels = return_labels if return_labels is not None else config.get('return_labels', False)
    return_labels = str2bool(return_labels)
    is_remote = is_remote if is_remote is not None else config.get('is_remote', True)
    is_remote = str2bool(is_remote)
    return output_dir, project, name, model_version, exporter_type, return_labels, is_remote


def create_metadata(inputs, outputs, sig_name, model_name, lengths_key=None, beam=None, return_labels=False, preproc='client'):
    meta = {
        'inputs': inputs,
        'outputs': outputs,
        'signature_name': sig_name,
        'metadata': {
            'exported_model': str(model_name),
            'exported_time': str(datetime.utcnow()),
            'return_labels': return_labels,
            'preproc': preproc,
        }
    }
    if lengths_key:
        meta['lengths_key'] = lengths_key
    if beam:
        meta['beam'] = beam

    return meta


def save_to_bundle(output_path, directory, assets=None, zip_results=False):
    """Save files to the exported bundle.

    :vocabs
    :vectorizers
    :labels
    :assets
    :output_path  the bundle output_path. vocabs, vectorizers know how to save themselves.
    """
    for filename in os.listdir(directory):
        if filename.startswith('vocabs') or \
           filename.endswith(".labels") or \
           filename.startswith('vectorizers'):
            shutil.copy(os.path.join(directory, filename), os.path.join(output_path, filename))

    if assets:
        asset_file = os.path.join(output_path, 'model.assets')
        write_json(assets, asset_file)

    if zip_results:
        zip_files(output_path, False)
        delete_old_copy(output_path)

def create_feature_exporter_field_map(feature_section, default_exporter_field='tokens'):
    feature_exporter_field_map = {}
    for feature_desc in feature_section:
        if feature_desc.get('exporter_field') is None:
            feature_exporter_field_map[feature_desc['name']] = default_exporter_field
        else:
            feature_exporter_field_map[feature_desc['name']] = feature_desc['exporter_field']
    return feature_exporter_field_map



from baseline.version import __version__
from baseline.utils import *
from baseline.vectorizers import *
from baseline.confusion import *
from baseline.data import *
from baseline.reader import *
from baseline.progress import *
from baseline.reporting import *
from baseline.model import *
from baseline.embeddings import *
from baseline.services import *
from baseline.train import *
logger = get_console_logger('baseline', env_key='BASELINE_LOG_LEVEL')
report_logger = get_console_logger('baseline.reporting', env_key='BASELINE_LOG_LEVEL')



import logging
import numpy as np
from baseline.utils import (
    exporter, optional_params, listify, register, import_user_module, read_json
)

__all__ = []
export = exporter(__all__)
logger = logging.getLogger('baseline')

BASELINE_MODELS = {}
BASELINE_LOADERS = {}


@export
@optional_params
def register_model(cls, task, name=None):
    """Register a function as a plug-in"""
    if name is None:
        name = cls.__name__

    names = listify(name)

    if task not in BASELINE_MODELS:
        BASELINE_MODELS[task] = {}

    if task not in BASELINE_LOADERS:
        BASELINE_LOADERS[task] = {}

    if hasattr(cls, 'create'):
        def create(*args, **kwargs):
            return cls.create(*args, **kwargs)
    else:
        def create(*args, **kwargs):
            return cls(*args, **kwargs)

    for alias in names:
        if alias in BASELINE_MODELS[task]:
            raise Exception('Error: attempt to re-define previously registered handler {} (old: {}, new: {}) for task {} in registry'.format(alias, BASELINE_MODELS[task], cls, task))

        BASELINE_MODELS[task][alias] = create

        if hasattr(cls, 'load'):
            BASELINE_LOADERS[task][alias] = cls.load
    return cls


@export
def create_model_for(activity, **kwargs):
    model_type = kwargs.get('type', kwargs.get('model_type', 'default'))
    creator_fn = BASELINE_MODELS[activity][model_type]
    logger.info('Calling model %s', creator_fn)
    input_ = kwargs.pop('features', None)
    output_ = kwargs.pop('labels', None)
    if output_ is not None:
        return creator_fn(input_, output_, **kwargs)
    return creator_fn(input_, **kwargs)


@export
def create_model(embeddings, labels, **kwargs):
    return create_model_for('classify', features=embeddings, labels=labels, **kwargs)


@export
def create_tagger_model(embeddings, labels, **kwargs):
    return create_model_for('tagger', features=embeddings, labels=labels, **kwargs)



BASELINE_SEQ2SEQ_ENCODERS = {}

@export
@optional_params
def register_encoder(cls, name=None):
    """Register a function as a plug-in"""
    return register(cls, BASELINE_SEQ2SEQ_ENCODERS, name, 'encoder')


BASELINE_SEQ2SEQ_DECODERS = {}

@export
@optional_params
def register_decoder(cls, name=None):
    """Register a function as a plug-in"""
    return register(cls, BASELINE_SEQ2SEQ_DECODERS, name, 'decoder')


BASELINE_SEQ2SEQ_ARC_POLICY = {}

@export
@optional_params
def register_arc_policy(cls, name=None):
    """Register a function as a plug-in"""
    return register(cls, BASELINE_SEQ2SEQ_ARC_POLICY, name, 'decoder')


@export
def create_seq2seq_decoder(tgt_embeddings, **kwargs):
    decoder_type = kwargs.get('decoder_type', 'default')
    Constructor = BASELINE_SEQ2SEQ_DECODERS.get(decoder_type)
    return Constructor(tgt_embeddings, **kwargs)


@export
def create_seq2seq_encoder(**kwargs):
    encoder_type = kwargs.get('encoder_type', 'default')
    Constructor = BASELINE_SEQ2SEQ_ENCODERS.get(encoder_type)
    return Constructor(**kwargs)


@export
def create_seq2seq_arc_policy(**kwargs):
    arc_type = kwargs.get('arc_policy_type', 'default')
    Constructor = BASELINE_SEQ2SEQ_ARC_POLICY.get(arc_type)
    return Constructor()


@export
def create_seq2seq_model(embeddings, labels, **kwargs):
    return create_model_for('seq2seq', embeddings, labels, **kwargs)


@export
def create_lang_model(embeddings, **kwargs):
    return create_model_for('lm', embeddings, None, **kwargs)


@export
def load_model_for(activity, filename, **kwargs):
    # Sniff state to see if we need to import things
    state = read_json('{}.state'.format(filename))
    if 'hub_modules' in state:
        for hub_module in state['hub_modules']:
            import_user_module(hub_module)
    # There won't be a module for pytorch (there is no state file to load).
    if 'module' in state:
        import_user_module(state['module'])
    # Allow user to override model type (for back compat with old api), backoff
    # to the model type in the state file or to default.
    # TODO: Currently in pytorch all models are always reloaded with the load
    # classmethod with a default model class. This is fine given how simple pyt
    # loading is but it could cause problems if a model has a custom load
    model_type = kwargs.get('type', kwargs.get('model_type', state.get('type', state.get('model_type', 'default'))))
    creator_fn = BASELINE_LOADERS[activity][model_type]
    logger.info('Calling model %s', creator_fn)
    return creator_fn(filename, **kwargs)


@export
def load_model(filename, **kwargs):
    return load_model_for('classify', filename, **kwargs)


@export
def load_tagger_model(filename, **kwargs):
    return load_model_for('tagger', filename, **kwargs)


@export
def load_seq2seq_model(filename, **kwargs):
    return load_model_for('seq2seq', filename, **kwargs)


@export
def load_lang_model(filename, **kwargs):
    return load_model_for('lm', filename, **kwargs)


@export
class ClassifierModel:
    """Text classifier

    Provide an interface to DNN classifiers that use word lookup tables.
    """
    task_name = 'classify'

    def __init__(self, *args, **kwargs):
        super().__init__()

    def save(self, basename):
        """Save this model out

        :param basename: Name of the model, not including suffixes
        :return: None
        """
        pass

    @classmethod
    def load(cls, basename, **kwargs):
        """Load the model from a basename, including directory

        :param basename: Name of the model, not including suffixes
        :param kwargs: Anything that is useful to optimize experience for a specific framework
        :return: A newly created model
        """
        pass

    def predict(self, batch_dict):
        """Classify a batch of text with whatever features the model can use from the batch_dict.
        The indices correspond to get_vocab().get('word', 0)

        :param batch_dict: This normally contains `x`, a `BxT` tensor of indices.  Some classifiers and readers may
        provide other features

        :return: A list of lists of tuples (label, value)
        """
        pass

    # deprecated: use predict
    def classify(self, batch_dict):
        logger.warning('`classify` is deprecated, use `predict` instead.')
        return self.predict(batch_dict)

    def get_labels(self):
        """Return a list of labels, where the offset within the list is the location in a confusion matrix, etc.

        :return: A list of the labels for the decision
        """
        pass


@export
class TaggerModel:
    """Structured prediction classifier, AKA a tagger

    This class takes a temporal signal, represented as words over time, and characters of words
    and generates an output label for each time.  This type of model is used for POS tagging or any
    type of chunking (e.g. NER, POS chunks, slot-filling)
    """
    task_name = 'tagger'

    def __init__(self):
        super().__init__()

    def save(self, basename):
        pass

    @staticmethod
    def load(basename, **kwargs):
        pass

    def predict(self, batch_dict):
        pass

    def get_labels(self):
        pass


@export
class LanguageModel(object):

    task_name = 'lm'

    def __init__(self):
        super().__init__()

    @staticmethod
    def load(basename, **kwargs):
        pass

    @classmethod
    def create(cls, embeddings, **kwargs):
        pass

    def predict(self, batch_dict, **kwargs):
        pass


@export
class EncoderDecoderModel(object):

    task_name = 'seq2seq'

    def save(self, model_base):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__()

    @staticmethod
    def load(basename, **kwargs):
        pass

    @classmethod
    def create(cls, src_embeddings, dst_embedding, **kwargs):
        pass

    def create_loss(self):
        pass

    def predict(self, source_dict, **kwargs):
        pass

    # deprecated: use predict
    def run(self, source_dict, **kwargs):
        logger.warning('`run` is deprecated, use `predict` instead.')
        return self.predict(source_dict, **kwargs)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_serving/apis/model_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from baseline.tensorflow_serving.apis import get_model_status_pb2 as tensorflow__serving_dot_apis_dot_get__model__status__pb2
from baseline.tensorflow_serving.apis import model_management_pb2 as tensorflow__serving_dot_apis_dot_model__management__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_serving/apis/model_service.proto',
  package='tensorflow.serving',
  syntax='proto3',
  serialized_pb=_b('\n+tensorflow_serving/apis/model_service.proto\x12\x12tensorflow.serving\x1a.tensorflow_serving/apis/get_model_status.proto\x1a.tensorflow_serving/apis/model_management.proto2\xe7\x01\n\x0cModelService\x12g\n\x0eGetModelStatus\x12).tensorflow.serving.GetModelStatusRequest\x1a*.tensorflow.serving.GetModelStatusResponse\x12n\n\x19HandleReloadConfigRequest\x12\'.tensorflow.serving.ReloadConfigRequest\x1a(.tensorflow.serving.ReloadConfigResponseB\x03\xf8\x01\x01\x62\x06proto3')
  ,
  dependencies=[tensorflow__serving_dot_apis_dot_get__model__status__pb2.DESCRIPTOR,tensorflow__serving_dot_apis_dot_model__management__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\370\001\001'))

_MODELSERVICE = _descriptor.ServiceDescriptor(
  name='ModelService',
  full_name='tensorflow.serving.ModelService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=164,
  serialized_end=395,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetModelStatus',
    full_name='tensorflow.serving.ModelService.GetModelStatus',
    index=0,
    containing_service=None,
    input_type=tensorflow__serving_dot_apis_dot_get__model__status__pb2._GETMODELSTATUSREQUEST,
    output_type=tensorflow__serving_dot_apis_dot_get__model__status__pb2._GETMODELSTATUSRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='HandleReloadConfigRequest',
    full_name='tensorflow.serving.ModelService.HandleReloadConfigRequest',
    index=1,
    containing_service=None,
    input_type=tensorflow__serving_dot_apis_dot_model__management__pb2._RELOADCONFIGREQUEST,
    output_type=tensorflow__serving_dot_apis_dot_model__management__pb2._RELOADCONFIGRESPONSE,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_MODELSERVICE)

DESCRIPTOR.services_by_name['ModelService'] = _MODELSERVICE

# @@protoc_insertion_point(module_scope)
